From iris.algebra Require Import base.
From igps.examples Require Import list_lemmas.

(* Hashtables are represented as partial association lists,
   with entries with key 0 considered to be empty. *)
Definition sublist {A} i j (l : list A) := firstn (j - i) (skipn i l).

Lemma sublist_nil {A} i (l : list A) : sublist i i l = nil.
Proof.
  unfold sublist.
  rewrite Nat.sub_diag; auto.
Qed.

Lemma sublist_0_cons {A} j (x : A) l : j > 0 ->
  sublist 0 j (x :: l) = x :: sublist 0 (j - 1) l.
Proof.
  unfold sublist; simpl.
  destruct j; [omega | simpl].
  rewrite !Nat.sub_0_r; auto.
Qed.

Class hash_fun := { size : nat; hash : Z -> nat; size_pos : size > 0;
  hash_range : forall i, 0 <= hash i < size }.

Section Hashtable.

Context {hf : hash_fun}.
Hint Resolve size_pos hash_range.

Definition rebase {A} (l : list A) i := sublist i (length l) l ++ sublist 0 i l.

Fixpoint index_of' (m : list Z) k :=
  match m with
  | [] => None
  | k1 :: rest => if decide (k1 = 0) then Some 0
                  else if decide (k1 = k) then Some 0
                  else option_map S (index_of' rest k)
  end.

Definition hlookup (m : list Z) (k : Z) :=
  option_map (fun i => (i + hash k) mod size) (index_of' (rebase m (hash k)) k).

Lemma index_of'_spec : forall k m, match index_of' m k with
  | Some i => i < length m /\ (nth i m 0%Z = k \/ nth i m 0%Z = 0%Z) /\
              Forall (fun x => x <> 0 /\ x <> k)%Z (sublist 0 i m)
  | None => ~k ∈ m /\ ~0%Z ∈ m
  end.
Proof.
  induction m; simpl; auto; intros.
  { split; intro; inversion H. }
  destruct (decide _).
  { subst; split; [omega|].
    rewrite sublist_nil; split; auto. }
  destruct (decide _).
  { subst; split; [omega|].
    rewrite sublist_nil; split; auto. }
  destruct (index_of' m k); simpl.
  - destruct IHm as (? & ? & ?).
    split; [omega|].
    split; auto.
    rewrite sublist_0_cons; last omega; constructor; auto; simpl.
    rewrite Nat.sub_0_r; auto.
  - split; intro; inversion H; subst; tauto.
Qed.

Lemma length_sublist : forall {A} i j (l : list A), i <= j <= length l ->
  length (sublist i j l) = j - i.
Proof.
  intros; unfold sublist.
  rewrite take_length drop_length.
  apply Min.min_l; omega.
Qed.

Lemma nth_sublist : forall {A} lo i hi l (d : A), i < hi - lo ->
  nth i (sublist lo hi l) d = nth (i + lo) l d.
Proof.
  intros; unfold sublist.
  rewrite nth_lookup lookup_take; last omega.
  rewrite lookup_drop.
  rewrite plus_comm.
  destruct (decide (i + lo < length l)).
  - destruct (nth_lookup_or_length l (i + lo) d); [by rewrite e | omega].
  - rewrite lookup_ge_None_2; last omega.
    rewrite nth_overflow; [done | omega].
Qed.

Lemma Forall_sublist_le : forall {A} (P : A -> Prop) i j l d
  (Hrangej : j <= length l) (Hi : ~P (nth i l d)) (Hj : Forall P (sublist 0 j l)),
  j <= i.
Proof.
  intros.
  destruct (decide (j <= i)); auto.
  eapply Forall_nth with (i0 := i) in Hj; [|rewrite length_sublist; omega].
  rewrite nth_sublist in Hj; try omega.
  rewrite Nat.add_0_r in Hj; done.
Qed.

Corollary Forall_sublist_first : forall {A} (P : A -> Prop) i j l d
  (Hrangei : 0 <= i <= length l) (Hi : Forall P (sublist 0 i l)) (Hi' : ~P (nth i l d))
  (Hrangej : 0 <= j <= length l) (Hj : Forall P (sublist 0 j l)) (Hj' : ~P (nth j l d)),
  i = j.
Proof.
  intros.
  apply Nat.le_antisymm; eapply Forall_sublist_le; eauto; omega.
Qed.

Lemma nth_elem : forall {A} (l : list A) i d, i < length l -> nth i l d ∈ l.
Proof.
  intros; eapply elem_of_list_lookup_2.
  destruct (nth_lookup_or_length l i d); [done | omega].
Qed.

Lemma index_of'_succeeds : forall k (m : list Z) i (Hi : i < length m)
  (Hnz : Forall (fun x => x <> 0 /\ x <> k)%Z (sublist 0 i m))
  (Hk : nth i m 0%Z = k \/ nth i m 0%Z = 0%Z), index_of' m k = Some i.
Proof.
  intros.
  pose proof (index_of'_spec k m).
  destruct (index_of' m k).
  - destruct H as (? & Hz & Hnz').
    f_equal; eapply Forall_sublist_first with (d := 0%Z); eauto; omega.
  - destruct H as (Hk' & Hz'); destruct Hk as [Hk | Hk]; [contradiction Hk' | contradiction Hz'];
      rewrite <- Hk; apply nth_elem; auto.
Qed.

Lemma length_rebase : forall {A} (m : list A) i, 0 <= i < length m ->
  length (rebase m i) = length m.
Proof.
  intros; unfold rebase.
  rewrite app_length !length_sublist; omega.
Qed.

Lemma Forall_forall_nth : forall {A} (P : A -> Prop) l d,
  Forall P l <-> forall i, i < length l -> P (nth i l d).
Proof.
  split; intros; [apply Forall_nth; auto|].
  induction l; auto; simpl in *.
  constructor.
  - apply (H 0); omega.
  - apply IHl; intros.
    apply (H (S i)); omega.
Qed.

Lemma nth_rebase : forall {A} (m : list A) i j d, i < length m -> j < length m ->
  nth i (rebase m j) d = nth ((i + j) mod length m) m d.
Proof.
  intros; unfold rebase.
  assert (length (sublist j (length m) m) = length m - j).
  { apply length_sublist; omega. }
  destruct (decide (i < length m - j)).
  - rewrite app_nth1; last omega.
    rewrite nth_sublist; try omega.
    rewrite Nat.mod_small; [done | omega].
  - rewrite app_nth2; last omega.
    rewrite nth_sublist; try omega.
    rewrite Nat.add_0_r H1.
    replace (i + j) with ((i + j - length m) + 1 * length m); last omega.
    rewrite Nat.mod_add; last omega.
    rewrite Nat.mod_small; last omega.
    f_equal; omega.
Qed.

Lemma hlookup_succeeds : forall (k : Z) (i : nat) i1 (T : list Z) (Hk : k ≠ 0) (Hi : (i < size)%nat)
  (Hi1 : (Z.to_nat (i1 mod size) = (i + hash k) mod size)%nat) (HT : length T = size)
  (Hfail : Forall (fun x => x ≠ 0%Z /\ x ≠ k) (sublist 0 i (rebase T (hash k))))
  (Hhit : nth (Z.to_nat (i1 mod size)) T 0 = k \/ nth (Z.to_nat (i1 mod size)) T 0 = 0),
  hlookup T k = Some (Z.to_nat (i1 mod size)).
Proof.
  intros.
  unfold hlookup.
  assert (0 <= hash k < length T) by (replace (length T) with size; apply hash_range).
  assert (length (rebase T (hash k)) = size).
  { rewrite length_rebase; auto; omega. }
  rewrite (index_of'_succeeds _ _ i); try done.
  unfold option_map; congruence.
  { omega. }
  { rewrite nth_rebase; try omega.
    rewrite Hi1 in Hhit; rewrite HT; auto. }
Qed.

Lemma sublist_split : forall {A} i j k (l : list A), i <= j ->
  sublist i (j + k) l = sublist i j l ++ sublist j (j + k) l.
Proof.
  intros; unfold sublist.
  replace (drop j l) with (drop (j - i) (drop i l)).
  rewrite take_take_drop.
  f_equal; omega.
  { rewrite drop_drop.
    f_equal; omega. }
Qed.

Lemma sublist_len_1 : forall {A} i (l : list A) d, i < length l ->
  sublist i (i + 1) l = nth i l d :: nil.
Proof.
  unfold sublist; intros.
  rewrite minus_plus.
  revert dependent l.
  induction i; destruct l; simpl; intros; try omega; auto.
  apply IHi; omega.
Qed.

Lemma map_insert : forall {A B} (f : A -> B) l i x, map f (<[i:=x]>l) = <[i:=f x]>(map f l).
Proof.
  induction l; auto; simpl; intros.
  destruct i; auto; simpl.
  rewrite IHl; auto.
Qed.

Lemma insert_over : forall {A} (l : list A) i x, length l <= i -> <[i:=x]>l = l.
Proof.
  induction l; auto; simpl; intros.
  destruct i; first omega; simpl.
  rewrite IHl; auto; omega.
Qed.

Lemma drop_insert_less : forall {A} n (l : list A) i x, n <= i ->
  drop n (<[i:=x]> l) = <[i - n := x]>(drop n l).
Proof.
  induction n; simpl; intros.
  { by rewrite Nat.sub_0_r. }
  destruct l; auto.
  destruct i; first omega; simpl.
  apply IHn; omega.
Qed.

Lemma sublist_insert_l : forall {A} (l : list A) i lo hi x, lo <= hi ->
  hi <= i < length l -> sublist lo hi (<[i:=x]>l) = sublist lo hi l.
Proof.
  intros; unfold sublist.
  rewrite drop_insert_less; last omega.
  apply take_insert; omega.
Qed.

Lemma sublist_insert_r : forall {A} (l : list A) i lo hi x, i < lo ->
  lo <= hi <= length l -> sublist lo hi (<[i:=x]>l) = sublist lo hi l.
Proof.
  intros; unfold sublist.
  rewrite drop_insert; [done | omega].
Qed.

Lemma take_insert_more : forall {A} n (l : list A) i x, i < n ->
  take n (<[i:=x]> l) = <[i:=x]>(take n l).
Proof.
  induction n; auto; simpl; intros.
  destruct l; auto.
  destruct i; auto; simpl.
  rewrite IHn; [done | omega].
Qed.

Lemma sublist_insert_m : forall {A} (l : list A) i lo hi x, lo <= i ->
  i < hi <= length l ->
  (sublist lo hi (<[i:=x]>l) = <[i - lo := x]>(sublist lo hi l)).
Proof.
  intros; unfold sublist.
  rewrite drop_insert_less; last omega.
  apply take_insert_more; omega.
Qed.

Lemma rebase_insert : forall {A} (m : list A) i j x, i < length m -> j < length m ->
  rebase (<[(i + j) mod length m := x]>m) j = <[i:=x]>(rebase m j).
Proof.
  intros; unfold rebase.
  rewrite insert_length.
  assert (length (sublist j (length m) m) = length m - j) as Hlen
    by (apply length_sublist; omega).
  destruct (decide (i < length m - j)%nat).
  - rewrite insert_app_l; last omega.
    rewrite Nat.mod_small; last omega.
    rewrite sublist_insert_m; try omega.
    rewrite sublist_insert_l; try omega.
    rewrite Nat.add_sub; done.
  - rewrite insert_app_r_alt; last omega.
    rewrite Hlen.
    replace ((i + j) mod length m) with (i - (length m - j)).
    rewrite sublist_insert_r; try omega.
    rewrite sublist_insert_m; try omega.
    rewrite Nat.sub_0_r; done.
    { rewrite <- (Nat.mod_small (_ - _) (length m)); last omega.
      rewrite <- (Nat.mod_add (_ - _) 1); last omega.
      simpl; f_equal; omega. }
Qed.

Corollary rebase_insert' : forall {A} (m : list A) (i j : nat) x, i < length m -> j < length m ->
  rebase (<[i:=x]>m) j = <[Z.to_nat ((i - j) mod length m) := x]>(rebase m j).
Proof.
  intros; rewrite <- rebase_insert; try done.
  rewrite <- (Nat2Z.id (_ mod _)), mod_Zmod; last omega.
  rewrite Nat2Z.inj_add Z2Nat.id; last by apply Z_mod_lt; omega.
  rewrite Zplus_mod_idemp_l Z.sub_simpl_r -mod_Zmod; last omega.
  rewrite Nat2Z.id Nat.mod_small; done.
  { apply Nat2Z.inj_lt.
    rewrite Z2Nat.id; apply Z_mod_lt; omega. }
Qed.

Lemma map_lookup : forall {A B} (f : A -> B) l i, map f l !! i = option_map f (l !! i).
Proof.
  induction l; auto; simpl; intros.
  destruct i; eauto.
Qed.

Lemma nth_insert : forall {A} (l : list A) i x j d, i < length l -> nth j (<[i := x]>l) d =
  if decide (j = i) then x else nth j l d.
Proof.
  intros; rewrite !nth_lookup; destruct (decide _).
  - subst; rewrite list_lookup_insert; auto.
  - rewrite list_lookup_insert_ne; auto.
Qed.

Lemma insert_in : forall {A} (l : list A) i x, i < length l -> x ∈ <[i:=x]>l.
Proof.
  induction l; simpl; intros; first omega.
  destruct i; simpl; constructor.
  apply IHl; omega.
Qed.

Lemma index_of'_insert : forall (m : list Z) (k : Z) i (Hrange : i < length m)
  (Hi : nth i m 0 = k \/ nth i m 0 = 0%Z),
  index_of' (<[i:=k]>m) k = index_of' m k.
Proof.
  intros.
  pose proof (index_of'_spec k m) as Hk; pose proof (index_of'_spec k (<[i:=k]>m)) as Hk'.
  destruct (index_of' m k), (index_of' (<[i:=k]>m) k); auto.
  - rewrite insert_length in Hk'.
    destruct Hk as (? & Hk & ?), Hk' as (? & Hz & Hall).
    destruct (decide (i < n0)).
    { rewrite sublist_insert_m in Hall; try omega.
      eapply Forall_forall in Hall as []; first done.
      apply insert_in.
      rewrite length_sublist; omega. }
    rewrite sublist_insert_l in Hall; try omega.
    rewrite nth_insert in Hz; last done.
    apply f_equal; destruct (decide _); subst; eapply Forall_sublist_first; eauto; simpl; try omega.
    + destruct Hi as [-> | ->]; tauto.
    + simpl; destruct Hk as [-> | ->]; tauto.
    + destruct Hz as [-> | ->]; tauto.
    + destruct Hk as [-> | ->]; tauto.
  - destruct Hk' as [Hk']; contradiction Hk'; apply insert_in; done.
  - destruct Hk as (Hk & Hz); destruct Hi as [Hi | Hi]; [contradiction Hk | contradiction Hz];
      rewrite <- Hi; apply nth_elem; omega.
Qed.

Lemma index_of'_insert2 : forall m k i k' (Hrange : i < length m)
  (Hi : index_of' m k <> Some i) (Hdiff : k' <> k /\ k' <> 0),
  index_of' (<[i:=k']>m) k = index_of' m k.
Proof.
  intros.
  pose proof (index_of'_spec k m) as Hk; pose proof (index_of'_spec k (<[i:=k']>m)) as Hk'.
  destruct (index_of' m k), (index_of' (<[i:=k']>m) k); auto.
  - rewrite insert_length in Hk'.
    destruct Hk as (? & Hk & ?), Hk' as (? & Hz & Hall).
    rewrite nth_insert in Hz; last done.
    destruct (decide _); first tauto.
    f_equal; apply Nat.le_antisymm.
    + eapply Forall_sublist_le; try eassumption; simpl; try omega.
      { rewrite insert_length; omega. }
      rewrite nth_insert; last done.
      destruct (decide _); first by subst.
      destruct Hk as [-> | ->]; tauto.
    + eapply Forall_sublist_le; try eassumption; simpl; try omega.
      destruct Hz as [-> | ->]; tauto.
  - destruct Hk as (? & Hk & ?).
    assert (nth n m 0 = nth n (<[i:=k']>m) 0) as Heq.
    { rewrite nth_insert; last done.
      destruct (decide _); by subst. }
    destruct Hk' as (Hk' & Hz'); destruct Hk as [Hk | Hk]; [contradiction Hk' | contradiction Hz'];
      rewrite -Hk Heq; apply nth_elem; rewrite insert_length; auto.
  - destruct Hk' as (Hrange' & Hz & ?).
    rewrite nth_insert in Hz; last done.
    destruct (decide _); first tauto.
    rewrite insert_length in Hrange'.
    destruct Hk as (Hk' & Hz'); destruct Hz as [Hz | Hz]; [contradiction Hk' | contradiction Hz'];
      rewrite <- Hz; apply nth_elem; auto.
Qed.

Lemma lookup_insert_same : forall m k i, hlookup m k = Some i -> length m = size ->
  i < length m -> hlookup (<[i:=k]>m) k = hlookup m k.
Proof.
  unfold hlookup; intros.
  pose proof (hash_range k).
  rewrite rebase_insert'; try omega.
  rewrite index_of'_insert; auto.
  - rewrite length_rebase; last omega.
    apply Nat2Z.inj_lt; rewrite Z2Nat.id; apply Z_mod_lt; omega.
  - pose proof (index_of'_spec k (rebase m (hash k))) as Hspec.
    destruct (index_of' (rebase m (hash k)) k); inversion H; subst.
    replace size with (length m).
    rewrite mod_Zmod; last omega.
    rewrite Nat2Z.inj_add Zminus_mod_idemp_l Z.add_simpl_r.
    destruct Hspec as (Hz & ? & ?).
    rewrite length_rebase in Hz; last omega.
    rewrite Zmod_small; last omega.
    rewrite Nat2Z.id; auto.
Qed.

Lemma lookup_insert_diff : forall m k i k', hlookup m k <> Some i -> length m = size ->
  i < length m -> k' <> k /\ k' <> 0 ->
  hlookup (<[i:=k']>m) k = hlookup m k.
Proof.
  unfold hlookup; intros.
  pose proof (hash_range k).
  rewrite rebase_insert'; try omega.
  rewrite index_of'_insert2; auto.
  - rewrite length_rebase; last omega.
    apply Nat2Z.inj_lt; rewrite Z2Nat.id; apply Z_mod_lt; omega.
  - intro X; rewrite X in H; contradiction H; simpl.
    replace size with (length m); apply f_equal.
    apply Nat2Z.inj; rewrite mod_Zmod; last omega.
    rewrite Nat2Z.inj_add Z2Nat.id.
    rewrite Zplus_mod_idemp_l Z.sub_simpl_r Zmod_small; omega.
    { apply Z_mod_lt; omega. }
Qed.

End Hashtable.
