From iris.program_logic Require Export weakestpre.
From iris.proofmode Require Import tactics.
From iris.algebra Require Import coPset.

From igps Require Export notation malloc escrows bigop.
From igps Require Import persistor viewpred weakestpre proofmode arith.
From igps.examples Require Import nat_tokens protocols.
From igps.gps Require Import singlewriter.

Import uPred.

(** This is the formalization of Circular Buffer from the original GPS paper.
  This proof is further simplified by using single writer protocols. **)

Section CircBuffer.

  Context (n: Z) (nGt1 : (1 < n)%Z).

  Local Notation wi := 0%Z.
  Local Notation ri := 1%Z.
  Local Notation b := 2%Z.

  Definition newBuffer : base.val :=
    λ: <>,
      let: "q" := malloc (#n + #b) in
        ["q" + #ri]_na <- #0 ;;
        ["q" + #wi]_na <- #0 ;;
        "q".

  Definition tryProd : base.val :=
    λ: "q" "x",
      let: "w" := ["q" + #wi]_at in
      let: "r" := ["q" + #ri]_at in
      let: "w'" := ("w" + #1) `mod` #n in
        if: "w'" = "r"
        then #0
        else
          ["q" + #b + "w"]_na <- "x" ;;
          ["q" + #wi]_at <- "w'" ;;
          #1.

  Definition tryCons : base.val :=
    λ: "q",
      let: "w" := ["q" + #wi]_at in
      let: "r" := ["q" + #ri]_at in
        if: "w" = "r"
        then #0
        else
          let: "x" := ["q" + #b + "r"]_na in
            ["q" + #ri]_at <- (("r" + #1) `mod` #n) ;;
            "x".

  Notation circBufR := (prodUR (coPset_disjUR) (coPset_disjUR)).

  Class cbG Σ := CbG { cb_tokG :> inG Σ circBufR }.
  Definition cbΣ : gFunctors := #[GFunctor (constRF circBufR)].

  Instance subG_cbΣ {Σ} : subG cbΣ Σ → cbG Σ.
  Proof. solve_inG. Qed.

  Section Interpretation.
    Context `{fG: !foundationG Σ, cirbG: !cbG Σ}.

    Set Default Proof Using "Type".
    Local Notation iProp := (iProp Σ).
    Local Notation vPred := (vProp Σ).

    Context (P: Z → vPred).

    Definition pTok (γ: gname) (i: nat)
      := own γ (CoPset {[Pos.of_succ_nat i]} , CoPset ∅).

    Definition pToks_from (γ: gname) (i: nat)
      := own γ (CoPset $ coPset_from_ex i, CoPset ∅).

    Definition cTok (γ: gname) (i: nat)
      := own γ (CoPset ∅, CoPset {[Pos.of_succ_nat i]}).

    Definition cToks_from (γ: gname) (i: nat)
      := own γ (CoPset ∅, CoPset $ coPset_from_ex i).

    Lemma pTok_exclusive γ i: ⎡pTok γ i⎤ ∗ ⎡pTok γ i⎤ ⊢ False : vProp Σ.
    Proof.
      iIntros "Tok". rewrite -embed_sep -own_op pair_op.
      iDestruct (own_valid with "Tok") as %[Valid _]. simpl in Valid.
      rewrite -> coPset_disj_valid_op, disjoint_singleton_l in Valid.
      exfalso. by apply Valid, elem_of_singleton.
    Qed.

    Lemma cTok_exclusive γ i: ⎡cTok γ i⎤ ∗ ⎡cTok γ i⎤ ⊢ False : vProp Σ.
    Proof.
      iIntros "Tok". rewrite -embed_sep -own_op pair_op.
      iDestruct (own_valid with "Tok") as %[_ Valid]. simpl in Valid.
      rewrite -> coPset_disj_valid_op, disjoint_singleton_l in Valid.
      exfalso. by apply Valid, elem_of_singleton.
    Qed.

    Lemma pToks_from_extract γ i :
      pToks_from γ i ⊢ pTok γ i ∗ pToks_from γ (S i).
    Proof.
      iIntros "Toks".
      rewrite /pToks_from /pTok. rewrite -own_op pair_op.
      rewrite !/op !/cmra_op /=.
      case (decide ({[Pos.of_succ_nat i]} ## coPset_from_ex (S i))) => [_|ND];
        last first.
        { exfalso. apply ND, coPset_from_disjoint. }
      iApply (own_mono with "Toks"). apply prod_included.
      by rewrite /= !coPset_disj_included union_idemp -coPset_from_insert.
    Qed.

    Lemma cToks_from_extract γ i :
      cToks_from γ i ⊢ cTok γ i ∗ cToks_from γ (S i).
    Proof.
      iIntros "Toks".
      rewrite -own_op pair_op !/op !/cmra_op /=.
      case (decide ({[Pos.of_succ_nat i]} ## coPset_from_ex (S i))) => [_|ND];
        last first.
        { exfalso. apply ND, coPset_from_disjoint. }
      iApply (own_mono with "Toks"). apply prod_included.
      by rewrite /= !coPset_disj_included union_idemp -coPset_from_insert.
    Qed.

    Lemma pToks_cToks_split γ i:
      own γ (CoPset $ coPset_from_ex i, CoPset $ coPset_from_ex i)
      ⊢ pToks_from γ i ∗ cToks_from γ i.
    Proof.
      iIntros "Toks".
      rewrite -own_op pair_op !/op !/cmra_op /=.
      case (decide (coPset_from_ex i ## ∅)) =>[_|ND]; last first.
        { exfalso. apply ND. set_solver+. }
      case (decide (∅ ## coPset_from_ex i)) =>[_|ND]; last first.
        { exfalso. apply ND. set_solver+. }
      iApply (own_mono with "Toks"). apply prod_included.
      rewrite /= !coPset_disj_included. set_solver+.
    Qed.

    Definition loc_at (q: loc) (i: nat) : loc :=
      (ZplusPos (b + i `mod` n) q).

    Definition PE (γ: gname) (q: loc) (i: nat): vPred
      := [es ⎡pTok γ i⎤ ⇝ own_loc (loc_at q i)]%I.

    Definition CE (γ: gname) (q: loc) (i: nat): vPred
      := [es ⎡cTok γ i⎤ ⇝ ∃ v, P v ∗ (loc_at q i) ↦ v]%I.

    Definition PP (γq: gname * loc) (i: nat) (v: Z): vPred
      := (□ (⌜v = i `mod` n⌝
            ∗ ∀ (j: nat), ⌜j < i⌝ → CE (γq.1) (γq.2) j))%Z%I.

    Definition CP (γq: gname * loc) (i: nat) (v: Z): vPred
      := (□ (⌜v = i `mod` n⌝
            ∗ ∀ (j: nat), ⌜j < i + n⌝ → PE (γq.1) (γq.2) j))%Z%I.

    Global Instance PP_persistent γ i v: Persistent (PP γ i v). apply _. Qed.
    Global Instance CP_persistent γ i v: Persistent (CP γ i v). apply _. Qed.

    Definition prodInt x γ : interpC Σ natProtocol :=
    (λ b _ s v, if b then True else PP (x,γ) s v)%I.
    Definition consInt x γ : interpC Σ natProtocol :=
    (λ b _ s v, if b then True else CP (x,γ) s v)%I.

    Lemma cb_escrows_list_alloc γ q (ni: nat):
      (ni ≤ n)%Z →
      ([∗ list] i ∈ seq (Z.to_nat b) ni, (ZplusPos (Z.of_nat i) q) ↦ ?)%I
      ={↑escrowN}=∗ ([∗ list] i ∈ seq 0 ni, PE γ q i)%I.
    Proof.
      move => Le. iIntros "OL". iApply fupd_big_sepL.
      rewrite (Nat2Z.id 2) -2!fmap_seq -list_fmap_compose big_sepL_fmap.
      iApply (big_sepL_mono with "OL").
      iIntros (? i H) "ol".
      apply lookup_seq_inv in H as [H1 H2]. simpl in H1. subst k.
      rewrite (_: Z.of_nat ((S ∘ S) i) = b + i `mod` n)%Z.
      - by iApply (escrow_alloc ⎡pTok γ i⎤%I with "ol").
      - rewrite 2!Nat2Z.inj_succ Zplus_succ_r_reverse Zmod_small; omega.
    Qed.

  End Interpretation.

  Section proof.
    Context `{cirbG : cbG Σ}.
    Local Notation iProp := (iProp Σ).
    Local Notation vProp := (vProp Σ).

    Context (P: Z → vProp)
            `{fG : foundationG Σ}
            `{gpG: !gpsG Σ natProtocol _}
            `{persG : persistorG Σ}.

    Set Default Proof Using "Type".

    Local Open Scope I.

    Definition Prod (q: loc) : vProp
      := ∃ (γ: gname) (i j: nat),
            ⌜i < j + (Z.to_nat n)⌝%Z
          ∗ [XP ZplusPos wi q in i | prodInt P γ q]_W
          ∗ [XP ZplusPos ri q in j | consInt γ q]_R
          ∗ ⎡pToks_from γ i⎤.

    Definition Cons (q: loc) : vProp
      := ∃ (γ: gname) (i j: nat),
            ⌜j ≤ i⌝%Z
          ∗ [XP ZplusPos wi q in i | prodInt P γ q]_R
          ∗ [XP ZplusPos ri q in j | consInt γ q]_W
          ∗ ⎡cToks_from γ j⎤.

    Local Close Scope I.

    Lemma newBuffer_spec:
      {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ }}}
          (newBuffer #())
      {{{ (q: loc), RET #q; Prod q ∗ Cons q }}}.
    Proof using nGt1.
      intros. iIntros "#[kI kP] Post".
      wp_seq. wp_op. wp_bind (malloc _).

      iApply (wp_mask_mono); first auto.
      iApply (malloc_spec with "kI []"); [omega|done|].

      iNext. iIntros (q) "oLs".

      iMod (own_alloc (CoPset $ coPset_from_ex 0, CoPset $ coPset_from_ex 0))
        as (γ) "Toks"; first done.

      iDestruct (pToks_cToks_split with "Toks") as "(Prods & Cons)".

      rewrite (_ : Z.to_nat (n + b) = S $ S $ Z.to_nat n); last first.
        { rewrite Z2Nat.inj_add; [|omega|omega].
          rewrite (_: Z.to_nat 2 = 2%nat); last auto. omega. }
      rewrite 2!big_sepL_cons -/seq.
      iDestruct "oLs" as "(owi & ori & oLs)".

      iAssert (|={⊤}=> ([∗ list] i ∈ seq 0 (Z.to_nat n), PE γ q i)%I)%I
        with "[oLs]" as ">#PEs".
      { iApply (fupd_mask_mono (↑escrowN)); first auto.
        iApply (cb_escrows_list_alloc with "oLs"). lia'. }

      wp_let. wp_op. wp_bind ([_]_na <- _)%E.

      iApply (GPS_SW_Init (consInt γ q) (0)%nat
                  ([XP ZplusPos 1 q in 0%nat | consInt γ q ]_W)
                  with "[$kI $ori]"); [done|done| |].
      { iSplitL ""; first done. iIntros "$". iModIntro.
        iSplitL ""; [auto|iNext]. iSplitL ""; iIntros "!#".
        - by rewrite Zmod_0_l.
        - simpl. iIntros (j). iIntros "%".
          iDestruct (big_sepL_lookup with "PEs") as "$".
          rewrite lookup_seq; [done|]. apply Nat2Z.inj_lt. lia'. }

      iNext. iIntros "cW".

      wp_seq. wp_op. wp_bind ([_]_na <- _)%E.


      iApply (GPS_SW_Init (prodInt P γ q) (0)%nat
                          ([XP ZplusPos 0 q in 0%nat | prodInt P γ q ]_W)
        with "[$kI $owi]"); [done|done|..].
      { iSplitL ""; first done. iIntros "$". iModIntro.
        iSplitL ""; [auto|iNext]. iSplitL ""; iIntros "!#".
        - by rewrite Zmod_0_l.
        - iIntros (j) "%". lia'. }

      iNext. iIntros "pW".
      wp_seq.
      iApply ("Post" with "[Prods Cons cW pW]").

      (* TODO: fix implicits or make it work with proofmode *)
      iDestruct (GPS_SW_Writer_Reader with "cW") as "[cW cR]".

      iDestruct (GPS_SW_Writer_Reader with "pW") as "[pW pR]".
      iSplitL "Prods pW cR"; iExists γ, O, O.
      - iFrame "Prods pW cR". iPureIntro. lia'.
      - by iFrame "Cons cW pR".
    Qed.

    Lemma tryProd_spec q v:
      {{{ ⎡PSCtx⎤ ∗ Prod q ∗ P v }}}
          (tryProd #q #v)
      {{{ (b: Z), RET (#b);
            Prod q ∗ (⌜b ≠ 0⌝%nat ∨ P v) }}}.
    Proof using nGt1.
      intros; iIntros "[#kI (Prod & P)] Post".
      iDestruct "Prod" as (γ i j0) "(% & pW & cR & pToks)".

      wp_lam.
      wp_lam. wp_bind ([_]_at)%E. wp_op.

      iApply (GPS_SW_Read_ex with "[$kI $pW]"); [done|done|..].
      { by iIntros "/= !> !> % #$ !>". }

      iNext. iIntros (w) "(pW & #CEs)".
      iDestruct "CEs" as "(% & CEs)".

      wp_let. wp_bind ([_]_at)%E. wp_op.

      iApply (GPS_SW_Read (consInt γ q) True%I (CP (γ,q))
                with "[$kI $cR]"); [done|done|..].
      { iSplitL.
        - iIntros (?) "_". iLeft. by iIntros "!>" (?) "(? & _)".
        - by iIntros "/= !> !> %% [_ #$] !>". }

      iNext. iIntros (j r) "(% & #cR2 & #PEs)".
      iDestruct "PEs" as "(% & PEs)".

      wp_let. wp_op. wp_op. wp_let.
      wp_op. case_bool_decide.
      - wp_if_true.
        iApply ("Post" with "[pToks pW P]").
        iSplitR "P"; last by iRight.
        iExists γ, i, j.
        iFrame "pToks pW cR2". iPureIntro.
        eapply Z.lt_le_trans; first exact H. by apply Zplus_le_compat_r, inj_le.

      - wp_if_false.
        wp_bind ([_]_na <- _)%E. wp_op. wp_op.

        assert (i + 1 < j + n)%Z. (* arithmetic *)
          { rewrite H0 H2 in H3. cbn in H1.
            assert (HLe : (Z.succ i ≤ j + n)%Z) by lia'.
            apply Zle_lt_or_eq in HLe as [Lt|Eq]; first auto.
            exfalso. apply H3.
            rewrite Z.add_mod_idemp_l; last omega.
            by rewrite -/(Z.succ _) Eq -{1}(Z.mul_1_l n) Z_mod_plus_full. }

        iDestruct (pToks_from_extract with "pToks") as "(iTok & pToks)".
        iAssert (PE γ q i)%I as "PEi".
        { iApply ("PEs" with "[]"); first auto. iPureIntro. omega. }

        iDestruct (escrow_apply ⎡pTok γ i⎤%I (own_loc (loc_at q i))
          with "[]") as "EA".
        { iIntros "pTok". by iApply (pTok_exclusive with "pTok"). }

        iSpecialize ("EA" with "[$PEi $iTok]"); first auto.

        iMod (fupd_mask_mono with "EA") as ">oL"; first auto.

        (* arithmetic *)
        rewrite (_: ZplusPos w (ZplusPos b q) = loc_at q i); last first.
        { rewrite H0 /loc_at /ZplusPos Z2Pos.id; [|lia']. f_equal. omega. }

        iApply (na_write with "[$kI $oL]"); [done|].

        iNext. iIntros "oL".

        iDestruct (escrow_alloc ⎡cTok γ i⎤%I (∃ v : Z, P v ∗ loc_at q i ↦ v)%I with "[oL P]")
          as "CEi".
          { iNext. iExists v. iFrame "P oL". }
        iMod (fupd_mask_mono with "CEi") as "#CEi"; first auto.

        wp_seq. wp_bind ([_]_at <- _)%E. wp_op.
        rewrite H0 Z.add_mod_idemp_l ; last omega.

        set Px : vProp := (∀ j1 : nat, ⌜j1 < (i + 1)⌝%Z → CE P γ q j1)%I.
        iApply (GPS_SW_ExWrite (prodInt P γ q) (i + 1)%nat
                        Px (λ s', [XP ZplusPos 0 q in s' | prodInt P γ q ]_W)%I
                  with "[$kI $pW]"); [|done|done|..].
        { rewrite Nat.add_1_r; apply Nat.le_succ_diag_r. }
        { iSplitR "".
          - iNext. iIntros (v5). iIntros "(#Px & _ & $)".
            iModIntro. iSplitL ""; first done. iNext.
            iSplitL ""; iIntros "!#"; [by rewrite Nat2Z.inj_add|].
            cbn. iIntros (k). iIntros (Hk).
            iApply ("Px"). iPureIntro. lia'.
          - iIntros (k ?). apply Zlt_succ_le, Zle_lt_or_eq in a as [Lt|Eq].
            + by iApply ("CEs" with "[%]").
            + apply Nat2Z.inj in Eq. by rewrite Eq. }

        iNext. iIntros "(_ & pW)".
        wp_seq.
        iApply ("Post" with "[pToks pW]").
        iSplitR ""; last by iLeft.
        iExists γ, (i + 1)%nat, j. iSplitL "".
        + iPureIntro. lia'.
        + rewrite -Nat.add_1_r. iFrame "pW pToks cR2".
    Qed.

    Lemma tryCons_spec q:
      {{{ ⎡PSCtx⎤ ∗ Cons q }}}
          (tryCons #q)
      {{{ (v: Z), RET (#v);
           Cons q ∗ (⌜v = 0⌝%Z ∨ P v) }}}.
    Proof using nGt1.
      intros; iIntros "[#kI Cons] Post".
      iDestruct "Cons" as (γ i0 j) "(% & pR & cW & cToks)".

      wp_lam. wp_bind ([_]_at)%E. wp_op.

      iApply (GPS_SW_Read (prodInt P γ q) True%I (PP P (γ,q))
                with "[$kI $pR]"); [done|done|..].
      { iSplitL.
        - iIntros (??). iLeft. iIntros "!>" (?) "(? & _)". by auto.
        - by iIntros "!> !> %% [_ #$] !>". }

      iNext. iIntros (i w) "(% & pR & #CEs)".
      iDestruct "CEs" as "(% & CEs)".

      wp_let. wp_bind ([_]_at)%E. wp_op.

      iApply (GPS_SW_Read_ex
                with "[$kI $cW]"); [done|done|..].
      { by iIntros "!> !> % #$ !>". }

      iNext. iIntros (r) "(cW & #PEs)".
      iDestruct "PEs" as "(% & PEs)".

      wp_let. wp_op. case_bool_decide.
      - wp_if_true.
        iApply ("Post").
        iSplitR ""; last by iLeft.
        iExists γ, i, j.
        iFrame "cToks pR cW". iPureIntro. etrans; eauto.
        by apply inj_le.

      - wp_if_false.
        wp_bind ([_]_na)%E.
        wp_op. wp_op.

        assert (j < i)%Z. (* arithmetic *)
          { assert (HLe: (j ≤ i)%Z). { etrans; [ exact H | by apply inj_le]. }
            apply Zle_lt_or_eq in HLe as [?|Eq]; first auto.
            exfalso. apply H3. by rewrite H1 H2 Eq. }

        iDestruct (cToks_from_extract with "cToks") as "(jTok & cToks)".

        iDestruct ("CEs" $! j with "[%]") as "CEj"; [done|].

        iDestruct (escrow_apply ⎡cTok γ j⎤%I (∃ v, P v ∗ loc_at q j ↦ v)%I
          with "[] [$CEj $jTok]") as "EA".
        { iIntros "cTok". iApply (cTok_exclusive with "cTok"). }

        iMod (fupd_mask_mono with "EA") as (v) "(P & >oL)"; [done|].

        (* arithmetic *)
        rewrite (_: ZplusPos r (ZplusPos b q) = loc_at q j); last first.
        { rewrite H2 /loc_at /ZplusPos Z2Pos.id; [|lia']. f_equal. omega. }

        iApply (na_read with "[$kI $oL]"); [done|].

        iNext. iIntros (x) "(% & oL)".

        iDestruct (escrow_alloc ⎡pTok γ (j + Z.to_nat n)%nat⎤%I
                               (own_loc (loc_at q (j + Z.to_nat n)%nat))
                     with "[oL]"
                  )
                  as "PEj".
          { rewrite (_: loc_at q (j + Z.to_nat n) = loc_at q j).
            - iFrame "oL".
            - rewrite /loc_at. do 2 f_equal.
              rewrite Nat2Z.inj_add Z2Nat.id; last omega.
              by rewrite -{1}(Z.mul_1_l n) Z_mod_plus_full. }
        iMod (fupd_mask_mono with "PEj") as "#PEj"; first auto.

        wp_let. wp_bind ([_]_at <- _)%E. wp_op.
        wp_op. wp_op.

        rewrite H2 Z.add_mod_idemp_l; last omega.

        set Px := (∀ j1 : nat, ⌜j1 < (j + 1) + n⌝%Z
                                    → PE γ q j1)%I.

        iApply (GPS_SW_ExWrite (consInt γ q) (j + 1)%nat
                       Px (λ s', [XP ZplusPos 1 q in s' | consInt γ q ]_W)%I
                  with "[$kI $cW]"); [|done|done|..].
        { rewrite Nat.add_1_r; apply Nat.le_succ_diag_r. }
        { iSplitR "".
          - iNext. iIntros (v5). iIntros "(#Px & _ & $)".
            iModIntro. iSplitL ""; first auto. iNext.
            iSplitL ""; iIntros "!#"; [by rewrite Nat2Z.inj_add|auto].
            cbn. iIntros (k). iIntros (Hk).
            iApply ("Px"). iPureIntro. lia'.
          - iIntros (k ?). rewrite Z.add_succ_l in a.
            apply Zlt_succ_le, Zle_lt_or_eq in a as [Lt|Eq].
            + by iApply ("PEs" with "[%]").
            + rewrite -(Nat2Z.id j) -Z2Nat.inj_add -?Eq ?Nat2Z.id;
                [done|..]; omega. }

        iNext. iIntros "(_ & cW)".
        wp_seq. subst x.
        iApply ("Post").
        iSplitR "P"; last by iRight.
        iExists γ, i, (j + 1)%nat. iSplitL ""; first (iPureIntro; lia').

        rewrite -Nat.add_1_r. iFrame "cW cToks pR".
    Qed.

  End proof.

End CircBuffer.
