From iris.program_logic Require Export weakestpre.
From iris.proofmode Require Import tactics.

From igps Require Import notation malloc escrows repeat protocols.
From igps.gps Require Import shared plain recursive.
From igps.examples Require Import unit_token.
Import uPred lang.base.

(** Simplified formalization of Michael-Scott queue
      thanks to single-writer protocols **)

Section MSQueue.
  Local Notation link := 0%nat.
  Local Notation data := 1%nat.
  Local Notation head := 0%nat.
  Local Notation tail := 1%nat.

  Definition newQueue : base.val :=
    (λ: <>,
      let: "s" := Alloc in
      ["s" + #link]_na <- #0  ;;
      let: "q" := malloc (#2) in
      ["q" + #head]_na <- ((int) "s") ;;
      ["q" + #tail]_na <- ((int) "s") ;;
      "q")%V.

  Definition findTail : base.val :=
    λ: "q",
      let: "n" := ["q" + #tail]_at in
      let: "n'" := [((loc)"n") + #link]_at in
      if: "n'" = #0
      then "n"
      else
        ["q" + #tail]_at <- "n'" ;;
        #0.

  Definition tryEnq : base.val :=
    λ: "q" "x",
      let: "n" := malloc (#2) in
      ["n" + #data]_na <- "x" ;;
      ["n" + #link]_na <- #0;;
      let: "t" := repeat (findTail "q") in
      if: (CAS (((loc)"t") + #link) #0 ((int)"n"))
      then ["q" + #tail]_at <- ((int)"n") ;; #1
      else #0.

  Definition tryDeq : base.val :=
    λ: "q",
      let: "s" := ["q" + #head]_at in
      let: "n" := [((loc)"s") + #link]_at in
      if: "n" = #0
      then #0
      else
        if: (CAS ("q" + #head) "s" "n")
        then [((loc)"n") + #data]_na
        else #0.

  Definition LinkState: Type := () + loc.

  Definition LSRel : relation LinkState :=
    λ a b, match a with
          | inl () => True
          | inr l =>
            match b with
            | inl () => False
            | inr l' => bool_decide (l = l')
            end
          end.

  Canonical LinkProtocol : protocolT := ProtocolT _ LSRel.

  Instance LinkPrt_facts : protocol_facts LinkProtocol.
  Proof. esplit.
    - move => [[] //|l /=]. by case_bool_decide.
    - move => [[] //|l1] [[] //|l2] [[] //|l3] /=.
      by repeat (case_bool_decide; auto; subst).
    - apply (populate (inl ())).
    - apply sum_countable.
  Qed.

  Local Notation Null := (inl () : pr_state LinkProtocol).
  Local Notation Linked l := (inr l : pr_state LinkProtocol).

  Section Interp.
    Context `{fG : foundationG Σ}
              `{msquG: !uTokG Σ}
              `{gpG: !gpsG Σ LinkProtocol _}
              `{agreeG : !gps_agreeG Σ LinkProtocol}
              `{persG : persistorG Σ}.

    Set Default Proof Using "Type".
    Local Notation iProp := (iProp Σ).
    Context (P : Z → vProp Σ).

    Definition DEQ (l: loc) (γ γ': gname) : vProp Σ
        := [es Tok γ ⇝ ∃ v, l ↦ v ∗ P v ∗ Tok γ']%I.

    Definition Link' : (gname -c> _) -> gname -c> interpC Σ LinkProtocol :=
      (fun F γ b l s z =>
        match s with
        | inl _ => ⌜z = 0⌝
        | inr l' => ⌜z = Z.pos l'⌝
                    ∗ ∃ γ', DEQ (ZplusPos data l') γ γ'
                            ∗ [PP (ZplusPos link l') in Null @ γ' | F γ']
        end)%I.

    Global Instance Link'_contractive:
        Contractive (Link').
    Proof.
      intros ? ? ? H ? ? ? s ?. rewrite /Link'.
      destruct s; [done|]. apply bi.sep_ne; [done|].
      apply bi.exist_ne => ?.
      apply bi.sep_ne; [done|].
      apply vGPS_PP_contractive. destruct n => //.
      by apply H.
    Qed.

    Definition Link : ∀ γ, interpC Σ LinkProtocol :=
      [rec Link'].

    Instance Link'_persistent γ F b l s z:
      Persistent ((Link' γ F b l s z)).
    Proof.
      rewrite /Link' /Persistent.
      iIntros "Link".
      repeat case_match; iDestruct "Link" as "#?"; done.
    Qed.

    Global Instance Link_persistent γ b l s z:
      Persistent (Link γ b l s z).
    Proof.
      rewrite (unf_int Link). (* TODO: why so slow? *)
      apply _.
    Qed.

    Definition Head : interpC Σ unitProtocol :=
      (λ b _ _ x,
        if b
        then
          ∃ γ, [PP (Z.to_pos (x + link)%Z) in Null @ γ | Link γ] ∗ Tok γ
        else
          ⌜0 < x⌝ ∗ ∃ γ, [PP (Z.to_pos (x + link)%Z) in Null @ γ | Link γ])%I.

    Definition Tail : interpC Σ unitProtocol :=
      (λ b _ _ x,
        if b then True
        else
          ⌜0 < x⌝ ∗ ∃ γ, [PP (Z.to_pos (x + link)%Z) in Null @ γ | Link γ])%I.

  End Interp.

  Section proof.
    Context `{fG : foundationG Σ}
              `{msquG: !uTokG Σ}
              `{gpG: !gpsG Σ LinkProtocol _}
              `{gpG2: !gpsG Σ unitProtocol _}
              `{agreeG : !gps_agreeG Σ LinkProtocol}
              `{agreeG2 : !gps_agreeG Σ unitProtocol}
              `{persG : persistorG Σ}.

    Set Default Proof Using "Type".

    Context (P : Z → vProp Σ).

    Definition Queue (q: loc) : vProp Σ :=
      ([PP (ZplusPos head q) in () | Head P] ∗ [PP (ZplusPos tail q) in () | Tail P])%I.

    Instance Queue_persistent q: Persistent (Queue q). Proof. apply _. Qed.

    Lemma newQueue_spec:
      {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ }}}
          (newQueue #())
      {{{ q, RET #q; Queue q }}}.
    Proof.
      intros. iIntros "#[kI kP] Post".
      wp_lam.
      iMod (own_alloc (Excl ())) as (γ) "Tok"; first done.
      wp_bind (Alloc).
      iApply (alloc with "kI"); [done|].
      iNext. iIntros (s) "os".
      wp_let. wp_bind ([_]_na <- _)%E. wp_op.
      rewrite (_: (ZplusPos 0 s) = s); last done.

      iApply (GPS_PP_Init (p:=γ) (Link P γ) Null
              with "[$kI $kP $os]"); [done|done|..].
      { rewrite !(unf_int (Link P)) /=. by iSplit. }
      iNext. iIntros "#Links".
      wp_seq. wp_bind (malloc _).

      iApply (wp_mask_mono); first auto.
      iApply (malloc_spec with "kI"); [omega|done|].
      iNext. iIntros (q) "oLs".
      rewrite big_sepL_cons /=.
      iDestruct "oLs" as "(oH & [oT _])". (* TODO: avoid emp *)

      wp_let. wp_bind ([_]_na <- _)%E. wp_op. wp_op.

      iApply (GPS_PP_Init_default (Head P) ()
               with "[$kI $kP] [$oH Tok]"); [done|done|..].
      { iSplitL "Tok".
        - iExists γ. by iFrame "Tok".
        - iSplit; first done. by iExists γ. }

      iNext. iIntros "Head".
      wp_seq. wp_bind ([_]_na <- _)%E. wp_op. wp_op.
      iApply (GPS_PP_Init_default (Tail P) ()
               with "[$kI $kP] [$oT]"); [done|done|..].
      { iSplitL ""; first done.
        iNext. iSplit; first done. by iExists γ. }

      iNext. iIntros "Tail".
      wp_seq.
      by iApply ("Post" with "[$Head $Tail]").
    Qed.

    Lemma findTail_spec q:
      {{{ ⎡PSCtx⎤ ∗ Queue q }}}
        findTail #q
      {{{ (n: Z), RET #n; ⌜n = 0⌝ ∨
            ⌜n > 0⌝ ∗ ∃ γ, [PP (Z.to_pos (n + link)%Z) in Null @ γ | Link P γ] }}}.
    Proof.
      intros. iIntros "[#kI #(oH & oT)] Post".
      wp_lam.
      wp_bind ([_]_at)%E. wp_op.

      set Q : pr_state unitProtocol → Z → vProp Σ
        := λ _ x, (⌜0 < x⌝
              ∗ (∃ γ : gname, [PP Z.to_pos (x + 0) in Null @ γ | Link P γ ]))%I.
      iApply (GPS_PP_Read (p:=()) (Tail P) True%I Q with "kI [$oT]");
        [done|done|..].
      { iSplitL "".
        - iNext. iIntros ([] ?). iLeft. by iIntros (x) "($ & _)".
        - rewrite monPred_objectively_unfold /=. iIntros (?????) "!> [_ #$] //". }
      iNext. iIntros ([] x) "(_ & _ & % & ox)".
      wp_seq. wp_bind ([_]_at)%E. wp_op. wp_op.
      rewrite (_ : ZplusPos 0 (Z.to_pos x) = Z.to_pos (x + 0)); last first.
      { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; last done. omega. }
      iDestruct "ox" as (γ) "#ox".
      set Q' : pr_state LinkProtocol → Z → vProp Σ
      := λ s z, (⌜z ≠ 0⌝ → (⌜0 < z⌝
                       ∗ ∃ γ', [PP Z.to_pos (z + link) in Null @ γ' | Link P γ']))%I.
      iApply (GPS_PP_Read (p:=γ) (Link P γ) True%I Q' with "kI [$ox]");
        [done|done|..].
      { iSplitL "".
        - iIntros (s') "_". iLeft. iIntros "!>"( v) "(LinkInt & _)".
          rewrite (unf_int (Link P)).
          iModIntro. iIntros "%". destruct s' as [|l'].
          + by iDestruct "LinkInt" as "%".
          + iDestruct "LinkInt" as "(% & LinkInt)".
            iDestruct "LinkInt" as (γ') "(_ & Link)".
            iSplitL ""; first (iPureIntro; lia).
            iExists γ'. subst v. iFrame "Link".
        - rewrite monPred_objectively_unfold /=. iIntros (????) "!> _ (_ & #$) //". }

      iNext. iIntros (s n') "(% & _ & Q)".
      wp_seq. wp_op; case_bool_decide.
      - wp_if_true.
        iApply ("Post" with "[ox]").
        iRight. iSplit; first (iPureIntro; omega).
        by iExists _.
      - wp_if_false. wp_bind ([_]_at <- _)%E. wp_op.
        iApply (GPS_PP_Write (p:=()) (Tail P) () (Q' s n') (λ _, True)%I
                with "[$kI Q oT]");
          [done|done|done|..].
        { iSplitR "Q"; last by iFrame "Q oT".
          iNext. iIntros "(Q & _)".
          iSplitR "Q"; first done. iModIntro. iSplitR "Q"; first done.
          by iApply ("Q" with "[%]"). }

        iNext. iIntros "oT2".
        wp_seq.
        iApply ("Post"); by iLeft.
    Qed.

    Lemma tryEnq_spec q x:
      {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ ∗ Queue q ∗ P x }}}
        tryEnq #q #x
      {{{ (z: Z), RET #z; ⌜z ≠ 0⌝ ∨ P x }}}.
    Proof.
      intros. iIntros "(#kI & #kP & #(oH & oT) & P) Post".
      wp_lam.
      wp_let.
      wp_bind (malloc _).

      iApply (wp_mask_mono); first auto.
      iApply (malloc_spec with "kI"); [omega|done|].
      iNext. iIntros (n) "oLs".
      rewrite big_sepL_cons /=.
      iDestruct "oLs" as "(ol & (od & _))". (* TODO: avoid emp *)
      wp_let. wp_bind ([_]_na <- _)%E. wp_op.
      iApply (na_write with "[$kI $od]"); [done|].
      iNext. iIntros "od". wp_seq.
      wp_bind([_]_na <- _)%E. wp_op.

      iMod (own_alloc (Excl ())) as (γ') "Tok"; first done.
      iApply (GPS_PP_Init (p:=γ')  (Link P γ') Null
               with "[$kI $kP $ol]"); [done|done|..].
      { rewrite !(unf_int (Link P)). by iSplit. }

      iNext. iIntros "Link'". wp_seq.
      wp_bind ((rec: "f" <> := _) _)%E.
      (iLöb as "IH" forall "kP oH oT").
      iApply wp_repeat; first auto.
      iApply (findTail_spec q with "[kI $oH $oT]"); first done.
      iNext. iIntros (t) "Cases".
      iExists t. iSplit; first done.
      iDestruct "Cases" as "[%|[% Link]]".
      - rewrite bool_decide_true; [|done]. do 2 iNext.
        iApply ("IH" with "P Post od Tok Link'"); iIntros "!#"; by iFrame "#".
      - iClear "IH". rewrite bool_decide_false; [|omega]. do 2 iNext.
        wp_seq. wp_bind (CAS _ _ _)%E.
        wp_op. wp_op. wp_op.
        rewrite (_: ZplusPos 0 (Z.to_pos t) = Z.to_pos (t + link)); last first.
        { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [cbv; case_match|]; omega. }

        iDestruct "Link" as (γ) "Link".
        set P': vProp Σ := (ZplusPos 1 n ↦ x ∗ P x ∗ Tok γ')%I.
        set Q : pr_state LinkProtocol → vProp Σ
          := λ s, (⌜s = Linked n⌝)%I.
        set R: pr_state LinkProtocol → Z → vProp Σ
          := λ _ _, (P x ∗ ZplusPos 1 n ↦ x)%I.
        iDestruct "Link'" as "#Link'".
        iApply (GPS_PP_CAS (p:=γ) (Link P γ) P' Q R
                  with "kI [P Tok od Link]");
          [done|done|..].
        { iFrame "Tok Link". iSplitR "P od"; last iSplitR "P od"; last first.
          - rewrite monPred_objectively_unfold /=.
            iIntros "{$P $od}" (????) "!> _ (_ & #$) //".
          - iNext. iIntros (??) "(_ & _ & _ & ? & ? & _)". by iFrame.
          - iNext. iIntros (s'). iIntros "(_ & LinkInt & P')".
            rewrite (unf_int (Link P)).
            destruct s' as [[]|l'].
            + iExists (Linked n).
              iSplitR "P'"; first done. iSplitR "P'"; first done.
              destruct (escrow_alloc (Tok γ)
                          (∃ v, ZplusPos data n ↦ v ∗ P v ∗ Tok γ'))%I as [EA].
              iDestruct (escrow_alloc (Tok γ) (∃ v, ZplusPos data n ↦ v ∗ P v ∗ Tok γ')%I with "[P']") as "DEQ".
              { iNext. iExists x. by iFrame "P'". }
              iMod (fupd_mask_mono with "DEQ") as "#DEQ"; first namespaces.solve_ndisj.
              iModIntro. rewrite !(unf_int (Link P)).
              iSplitL ""; (iSplitL ""; first done);
                iExists γ'; iFrame "DEQ"; iFrame "Link'".
            + iDestruct "LinkInt" as "(% & _)". exfalso. lia. }

        iNext. iIntros (s' [] v) "(_ & Link & IF)".
        + wp_if_true.
          wp_bind ([_]_at <- _)%E. wp_op. wp_op.
          iApply (GPS_PP_Write (p:=()) (Tail P) () True%I (λ _, True)%I
                with "[$kI oT]");
                [done|done|done|..].
          { iFrame "oT". iSplitL; last done.
            iIntros "!> _". iModIntro. iSplitL; first done.
            iSplitL; first done. iSplitL.
            - iPureIntro. lia.
            - by iExists γ'. }
          iNext. iIntros "oT2".
          wp_seq.
          iApply ("Post"); by iLeft.
        + wp_if_false.
          iApply ("Post"); iRight.
          by iDestruct "IF" as "[$ ?]".
    Qed.

    Lemma tryDeq_spec q:
      {{{ ⎡PSCtx⎤ ∗ Queue q }}}
        tryDeq #q
      {{{ x, RET #x; ⌜x = 0⌝ ∨ P x }}}.
    Proof.
      intros. iIntros "[#kI #(oH & oT)] Post".
      wp_lam. wp_bind ([_]_at)%E. wp_op.
      set Q : pr_state unitProtocol → Z → _
        := λ _ x, (⌜0 < x⌝%Z
              ∗ (∃ γ : gname, [PP Z.to_pos (x + 0) in Null @ γ | Link P γ ]))%I.
      iApply (GPS_PP_Read (p:=()) (Head P) True%I Q with "kI [$oH]");
        [done|done|..].
      { iSplitL.
        - iIntros ([]) "_". iLeft. by iIntros "!>" (x) "($ & _)".
        - rewrite monPred_objectively_unfold /=. iIntros "!>" (????) "_ (_ & #$) //". }

      iNext. iIntros ([] s) "(_ & oH2 & % & os)".
      iDestruct "os" as (γ) "os".

      wp_seq. wp_bind ([_]_at)%E.
      wp_op. wp_op.
      rewrite (_: ZplusPos 0 (Z.to_pos s) = Z.to_pos (s + 0));
        last first. { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; omega. }

      set Q' : pr_state LinkProtocol → Z → _
        := λ s' n,
          (⌜0 ≠ n⌝
              → (⌜0 < n⌝ ∗ ∃ γ', DEQ P (Z.to_pos (n + data)) γ γ'
                      ∗ [PP Z.to_pos (n + link) in Null @ γ' | Link P γ']))%I.
      iApply (GPS_PP_Read (p:=γ) (Link P γ) True%I Q' with "kI [$os]");
        [done|done|..].
      { iSplitL.
        - iIntros (s'). iIntros "_". iLeft.
          iIntros (n). iIntros "!> (LinkInt & _)".
          rewrite (unf_int (Link P)).
          iModIntro. iIntros "%".
          destruct s' as [|l'].
          + by iDestruct "LinkInt" as "%".
          + iDestruct "LinkInt" as "(% & LinkInt)".
            iSplitR "LinkInt"; first (iPureIntro; lia).
            iDestruct "LinkInt" as (γ') "(DEQ & Link)".
            iExists γ'. subst.
            rewrite (_: Z.to_pos (Z.pos l' + 1) = ZplusPos 1 l'); last first.
            { rewrite /ZplusPos. f_equal. omega. }
            iFrame "DEQ Link".
        - rewrite monPred_objectively_unfold /=. iIntros "!>" (????) "_ [_ #$] //". }

      iNext. iIntros (s' n) "(_ & #os & Q')".
      wp_seq. wp_op. case_bool_decide.
      - wp_if_true.
        iApply ("Post"); by iLeft.
      - wp_if_false. wp_bind (CAS _ _ _)%E. wp_op.
        iDestruct ("Q'" with "[%]") as "(% & Q')"; [done|].
        iDestruct "Q'" as (γ') "(#DEQ & #on)".

        set Q1 : pr_state unitProtocol → _
          := λ _, (▷ ∃ x, (Z.to_pos (n + data)) ↦ x ∗ P x)%I.
        iApply (GPS_PP_CAS (p:=()) (Head P) True%I Q1 (λ _ _, True)%I
                  with "kI []"); [done|done|..].
        { iSplitL; last iSplitL; last first.
          - rewrite monPred_objectively_unfold /=.
            iIntros "{$oH} !>" (?????) "[_ #$] //".
          - iNext. by iIntros (??) "_".
          - iNext. iIntros ([]).
            iIntros "(_ & Head & _)".
            iDestruct "Head" as (γ2) "(os2 & Tok)".
            iAssert (⌜γ2 = γ⌝)%I with "[#]" as %?.
            { iApply (GPS_PPs_param_agree γ2 γ with "[$os2 $os]"). }
            subst γ2.
            iDestruct ((escrow_apply (Tok γ)
                        (∃ x, (Z.to_pos (n + data)) ↦ x ∗ P x ∗ Tok γ'))%I
                      with "[]") as "EA".
            { iIntros "[? ?]". iApply unit_tok_exclusive. iFrame. }
            iSpecialize ("EA" with "[$DEQ $Tok]").
            iDestruct (fupd_mask_mono with "EA") as "EA";
            [|iMod "EA" as (x) "(od & P & >Tok)"]; [namespaces.solve_ndisj|].
            iModIntro. iExists (). iSplitL ""; first auto.
            iSplitL "od P".
            + iExists x. iFrame "od P".
            + iSplitL "Tok".
              * iExists γ'. iFrame "on Tok".
              * iSplitL; first done. iExists γ'. iFrame "on". }

        iNext. iIntros ([] [] v) "(_ & _ & IF)".
        + wp_if_true. wp_op. wp_op.
          rewrite (_ : ZplusPos 1 (Z.to_pos n) = Z.to_pos (n + 1)); last first.
          { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|auto]. }
          iDestruct "IF" as (x) "[ox P]".
          iApply (na_read with "[$kI $ox]"); [done|].
          iNext. iIntros (z) "(% & ox)".
          subst z.
          iApply ("Post" with "[P]"); by iRight.
        + wp_if_false.
          iApply ("Post"); by iLeft.
    Qed.
  End proof.
End MSQueue.
