From iris.program_logic Require Export weakestpre.
From iris.proofmode Require Import tactics.

From igps Require Import notation malloc na proofmode weakestpre.

Import uPred.

(** An SC stack that is built up by non-atomics.
    This is used to batch deallocations in RCU. **)

Section SCStack.

  Open Scope Z_scope.
  Local Notation next := 0.
  Local Notation data := 1.

  Definition newStack : base.val :=
    λ: <>,
      let: "s" := Alloc in
      ["s"]_na <- #0  ;; "s".

  Definition push : base.val :=
    λ: "s" "x",
      let: "n" := malloc (#2) in
      ["n" + #data]_na <- "x" ;;
      let: "h" := ["s"]_na in
      ["n" + #next]_na <- "h";;
      ["s"]_na <- ((int)"n").

  Definition pop : base.val :=
    λ: "s",
      let: "h" := ["s"]_na in
      if: "h" = #0
      then #0
      else
        let: "v" := [((loc)"h") + #data]_na in
        let: "h'" := [((loc)"h") + #next]_na in
        ["s"]_na <- "h'" ;;
        Dealloc ((loc)"h" + #data) ;;
        Dealloc ((loc)"h" + #next) ;;
        "v".

  Section proof.
    Context `{fG : foundationG Σ}.
    Set Default Proof Using "Type".
    Notation vPred := (vProp Σ).

    Context (P: Z → vPred).

    Definition SCStack': (loc -c> list Z -c> vProp Σ)
                             -c> loc -c> list Z -c> vProp Σ
    := (fun F l A =>
        ∃ l', ZplusPos next l ↦ l' ∗
        match A with
        | nil => ⌜l' = 0⌝
        | v::A' => ⌜0 < l'⌝ ∗ Z.to_pos (l' + data) ↦ v ∗ P v
                              ∗ ▷ F (Z.to_pos l') A'
        end)%I.

    Instance SCStack'_inhabited: Inhabited (loc -c> list Z -c> vProp Σ)
      := populate (λ _ _, True)%I.

    Instance SCStack'_contractive:
        Contractive (SCStack').
    Proof.
      intros ? ? ? H ? A.
      repeat (apply bi.exist_ne => ?).
      apply bi.sep_ne; [done|].
      destruct A as [|v A]; [done|].
      repeat (apply bi.sep_ne; [done|]).
      apply later_contractive. destruct n => //. by apply (H).
    Qed.

    Definition SCStack := fixpoint (SCStack').

    Lemma newStack_spec:
      {{{ ⎡PSCtx⎤ }}}
         newStack #()
      {{{ (s: loc), RET #s; SCStack s nil }}}.
    Proof.
      iIntros (Φ) "#kI Post".
      wp_seq. wp_bind Alloc.
      wp_apply (alloc with "kI"); [done|].
      iIntros (x) "os".
      wp_seq.
      wp_bind ([_]_na <-_)%E.
      wp_apply (na_write with "[$kI $os]"); [done|].
      iIntros "os".
      wp_seq.
      iApply ("Post" with "[os]").
      rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
      iExists _. by iFrame "os".
    Qed.

    Lemma push_spec s v A:
      {{{ ⎡PSCtx⎤ ∗ SCStack s A ∗ P v }}}
         push #s #v
      {{{ RET #(); SCStack s (v :: A) }}}.
    Proof.
      intros. iIntros "[#kI (Stack & P)] Post".

      wp_lam. wp_let.
      wp_bind (malloc _).

      iApply (wp_mask_mono); first auto.
      wp_apply (malloc_spec with "kI []"); [omega|done|].
      iIntros (n) "oLs".
      iDestruct "oLs" as "(ol & od & _)".
      wp_let. wp_bind ([_]_na <- _)%E. wp_op.
      wp_apply (na_write with "[$kI $od]"); [done|].
      iIntros "od".
      wp_seq. wp_bind ([_]_na)%E.

      rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
      iDestruct "Stack" as (h) "[oH oT]".
      wp_apply (na_read with "[$kI $oH]"); [done|].
      iIntros (z) "[% oH]". subst z.
      wp_seq. wp_bind ([_]_na <- _)%E. wp_op.
      wp_apply (na_write with "[$kI $ol]"); [done|].
      iIntros "ol".
      wp_seq. wp_op.
      wp_apply (na_write with "[$kI $oH]"); [done|].

      iIntros "oH".
      iApply ("Post").

      rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
      iExists _. iFrame "oH P". iSplitL ""; first (iPureIntro; lia).
      iSplitL "od".
      - rewrite (_: Z.to_pos (Z.pos n + data) = ZplusPos data n); first done.
        rewrite /ZplusPos. f_equal. omega.
      - iNext. rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
        iExists h. iFrame "oT ol".
    Qed.

    Lemma pop_spec s A:
      {{{ ⎡PSCtx⎤ ∗ SCStack s A  }}}
         pop #s
      {{{ (z: Z), RET #z;
           match A with
           | nil => ⌜z = 0⌝ ∗ SCStack s A
           | v :: A' =>  ⌜z = v⌝ ∗ SCStack s A' ∗ P v
           end }}}.
    Proof.
      intros. iIntros "[#kI Stack] Post".
      wp_lam.
      wp_bind ([_]_na)%E.
      rewrite {1}/SCStack (fixpoint_unfold (SCStack') _ _).
      iDestruct "Stack" as (h) "[oH oT]".
      wp_apply (na_read with "[$kI $oH]"); [done|].
      iIntros (z) "[% oH]". subst z.
      wp_seq.
      destruct A as [|v A'].
      - iDestruct "oT" as "%". subst h.
        wp_op. wp_if.
        iApply ("Post").
        iSplitL ""; first done.
        rewrite /SCStack (fixpoint_unfold (SCStack') _ _). iExists _. by iFrame "oH".
      - iDestruct "oT" as "(% & od & P & oT)".
        wp_op. case_bool_decide; first omega.
        wp_if.
        wp_bind ([_]_na)%E. wp_op. wp_op.

        rewrite (_: ZplusPos data (Z.to_pos h) = Z.to_pos (h + 1)); last first.
        { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|auto]. }
        wp_apply (na_read with "[$kI $od]"); [done|].

        iIntros (z) "[% od]". subst z.
        wp_seq.
        wp_bind ([_]_na)%E. wp_op. wp_op.

        rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
        iDestruct "oT" as (h') "[oP oT]".
        wp_apply (na_read with "[$kI $oP]"); [done|].
        iIntros (z) "[% oP]". subst z.
        wp_seq. wp_bind ([_]_na <- _)%E.
        wp_apply (na_write with "[$kI $oH]"); [done|].
        iIntros "oH".
        wp_seq.
        wp_bind (Dealloc _). wp_op. wp_op.

        iApply (wp_mask_mono); first auto.
        rewrite (_: ZplusPos data (Z.to_pos h) = Z.to_pos (h + data)); last first.
        { rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|auto]. }
        wp_apply (dealloc with "[$kI $od]").

        iIntros "_".
        wp_seq.
        wp_bind (Dealloc _). wp_op. wp_op.
        iApply (wp_mask_mono); first auto.
        wp_apply (dealloc with "[$kI $oP]").
        iIntros "_".
        wp_seq.

        iApply ("Post").
        iSplitL ""; first done. iFrame "P".
        rewrite /SCStack (fixpoint_unfold (SCStack') _ _).
        iExists _. by iFrame.
    Qed.
  End proof.
End SCStack.
