From iris.algebra Require Export local_updates frac agree.
From iris.algebra Require Import proofmode_classes.
From iris.base_logic Require Import base_logic.
Set Default Proof Using "Type".

(* I don't think this can be built out of existing components. The only way to combine frac with
   something else is using prodR, which applies the operation pointwise. *)
Record snap (A : Type) := Snap { master : option (frac * agree A); snapshot : option A }.
Add Printing Constructor snap.
Arguments Snap {_} _ _.
Arguments master {_} _.
Arguments snapshot {_} _.
Instance: Params (@Snap) 1.
Instance: Params (@master) 1.
Instance: Params (@snapshot) 1.
Definition Snapshot {A} (a : A) := Snap None (Some a).
Definition Master {A} `{Unit A} k (a : A) := Snap (Some (k, to_agree a)) (Some ε).

(* COFE *)
Section cofe.
Context {A : ofeT}.
Implicit Types x y : snap A.

Instance snap_equiv : Equiv (snap A) := λ x y,
  master x ≡ master y ∧ snapshot x ≡ snapshot y.
Instance snap_dist : Dist (snap A) := λ n x y,
  master x ≡{n}≡ master y ∧ snapshot x ≡{n}≡ snapshot y.

Global Instance Snap_ne : NonExpansive2 (@Snap A).
Proof. by split. Qed.
Global Instance Snap_proper : Proper ((≡) ==> (≡) ==> (≡)) (@Snap A).
Proof. by split. Qed.
Global Instance master_ne: NonExpansive (@master A).
Proof. by destruct 1. Qed.
Global Instance master_proper : Proper ((≡) ==> (≡)) (@master A).
Proof. by destruct 1. Qed.
Global Instance own_ne : NonExpansive (@snapshot A).
Proof. by destruct 1. Qed.
Global Instance own_proper : Proper ((≡) ==> (≡)) (@snapshot A).
Proof. by destruct 1. Qed.

Definition snap_ofe_mixin : OfeMixin (snap A).
Proof. by apply (iso_ofe_mixin (λ x, (master x, snapshot x))). Qed.
Canonical Structure snapC := OfeT (snap A) snap_ofe_mixin.

(*Global Instance snap_cofe `{Cofe A} : Cofe snapC.
Proof.
  apply (iso_cofe (λ y : _ * _, Snap (y.1) (y.2))
    (λ x, (master x, snapshot x))); by repeat intro.
Qed.*)

Global Instance Snap_discrete a b :
  Discrete a → Discrete b → Discrete (Snap a b).
Proof. by intros ?? [??] [??]; split; apply: discrete. Qed.
Global Instance snap_discrete : OfeDiscrete A → OfeDiscrete snapC.
Proof. intros ? [??]; apply _. Qed.
End cofe.

Arguments snapC : clear implicits.

Class Join {A : ofeT} {D : OfeDiscrete A} {L : LeibnizEquiv A} {U : Unit A} (R : relation A) := {
  join_order :> PartialOrder R; unit_min : forall a, R ε a;
  join : A -> A -> option A;
  join_assoc : forall a b c d e, join a b = Some c -> join c d = Some e ->
    exists c', join a d = Some c' /\ join b c' = Some e; join_comm :> Comm eq join;
  join_le : forall a b c, join a b = Some c -> R a c;
  le_join : forall a b, R a b -> join a b = Some b }.

(* can this be done automatically with fmap? *)
Definition lift_R `{J : Join} a b := match a, b with Some a, Some b => R a b | _, _ => False end.
Definition lift_join `{J : Join} a b :=
  match a, b with Some a, Some b => join a b | _, _ => None end.

Lemma join_refl `{J : Join} (a : A) : join a a = Some a.
Proof.
  apply le_join; reflexivity.
Qed.

Lemma lift_join_le `{J : Join} : forall a b c, lift_join a b = Some c -> lift_R a (Some c).
Proof.
  intros [|] [|] ?; simpl; try discriminate; apply join_le.
Qed.

Lemma le_lift_join `{J : Join} : forall a b, lift_R a b -> lift_join a b = b.
Proof.
  intros [|] [|]; try contradiction; simpl; apply le_join.
Qed.

Lemma join_unit `{J : Join} : forall a, join ε a = Some a.
Proof.
  intro; apply le_join, unit_min.
Qed.

Instance lift_join_comm `{J : Join} : Comm eq lift_join.
Proof.
  intros [|] [|]; auto; apply join_comm.
Qed.

Instance lift_join_assoc `{J : Join} : Assoc eq lift_join.
Proof.
  intros [x|] [y|] [z|]; auto; simpl.
  destruct (join y z) eqn: Hj1, (join x y) eqn: Hj2; auto; simpl.
  + destruct (join x o) eqn: Hj3.
    * rewrite join_comm in Hj3; destruct (join_assoc _ _ _ _ _ Hj1 Hj3) as (? & Hj' & Hj'').
      rewrite join_comm in Hj'; rewrite Hj' in Hj2; inversion Hj2; subst.
      rewrite join_comm in Hj''; auto.
    * destruct (join o0 z) eqn: Hj4; auto.
      rewrite join_comm in Hj2; destruct (join_assoc _ _ _ _ _ Hj2 Hj4) as (? & Hj' & Hj'').
      rewrite Hj' in Hj1; inversion Hj1; subst.
      rewrite Hj'' in Hj3; discriminate.
  + destruct (join x o) eqn: Hj3; auto.
    rewrite join_comm in Hj3; destruct (join_assoc _ _ _ _ _ Hj1 Hj3) as (? & Hj' & ?).
    rewrite join_comm in Hj'; rewrite Hj' in Hj2; discriminate.
  + destruct (join o z) eqn: Hj3; auto.
    rewrite join_comm in Hj2; destruct (join_assoc _ _ _ _ _ Hj2 Hj3) as (? & Hj' & ?).
    rewrite Hj' in Hj1; discriminate.
  + destruct (join x y); auto.
Qed.

Lemma lift_join_idemp `{J : Join} : forall a, lift_join a a = a.
Proof.
  destruct a; auto; apply join_refl.
Qed.

(* CMRA *)
Section cmra.
Context `{J : Join A R}.
Implicit Types a b : A.
Implicit Types x y : snap A.

Instance snap_valid : Valid (snap A) := λ x, match snapshot x with None => False | Some b =>
  match master x with
  | Some (k, a) => ✓ k ∧ R b (hd ε (agree_car a)) ∧ ✓ a
  | None => True
  end end.
Global Arguments snap_valid !_ /.
Instance snap_validN : ValidN (snap A) := λ n x, match snapshot x with None => False | Some b =>
  match master x with
  | Some (k, a) => ✓{n} k ∧ R b (hd ε (agree_car a)) ∧ ✓{n} a
  | None => True
  end end.
Global Arguments snap_validN _ !_ /.
Instance snap_pcore : PCore (snap A) := λ x, Some {| master := None; snapshot := snapshot x |}.
Instance snap_op : Op (snap A) := λ x y,
  Snap (master x ⋅ master y) (lift_join (snapshot x) (snapshot y)).

Definition snap_valid_eq :
  valid = λ x, match snapshot x with None => False | Some b => match master x with
               | Some (k, a) => ✓ k ∧ R b (hd ε (agree_car a)) ∧ ✓ a
               | None => True
               end end := eq_refl _.
Definition snap_validN_eq :
  validN = λ n x, match snapshot x with None => False | Some b => match master x with
                  | Some (k, a) => ✓{n} k ∧ R b (hd ε (agree_car a)) ∧ ✓{n} a
                  | None => True
                  end end := eq_refl _.

Lemma master_validN n x : ✓{n} x → ✓{n} master x.
Proof. destruct x as [[[]|]]; try done; simpl.
  rewrite snap_validN_eq; simpl.
  destruct snapshot0; last contradiction; constructor; tauto.
Qed.

Lemma snap_validN_2 n k a b : ✓{n} (Master k a ⋅ Snapshot b) ↔ ✓{n} k ∧ R b a.
Proof. rewrite snap_validN_eq /= join_unit.
  repeat split; tauto.
Qed.

Lemma master_valid  x : ✓ x → ✓ master x.
Proof. destruct x as [[[]|]]; try done; simpl.
  rewrite snap_valid_eq; simpl.
  destruct snapshot0; last contradiction; constructor; tauto.
Qed.

Lemma init_valid a : ✓ (Master 1 a ⋅ Snapshot a).
Proof.
  hnf; simpl.
  rewrite join_unit; done.
Qed.

Lemma snap_cmra_mixin : CmraMixin (snap A).
Proof.
  apply cmra_total_mixin.
  - eauto.
  - intros n x y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'; try done.
    rewrite <- discrete_iff, leibniz_equiv_iff in *; try apply _.
    rewrite Hy'; auto.
  - by intros n y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'.
  - intros n [x a] [y b] [Hx Ha]; simpl in *. rewrite !snap_validN_eq; simpl.
    rewrite <- discrete_iff, leibniz_equiv_iff in Ha; last apply _; subst.
    destruct b; last done.
    destruct Hx as [[][] Hx|]; first destruct Hx; intros ?; ofe_subst; auto.
    assert (✓{n} o3) by (ofe_subst; tauto).
    assert (∀ a b, a ∈ agree_car o3 → b ∈ agree_car o3 → a ≡{n}≡ b) as Hvalid
      by (rewrite <- agree_validN_def; done).
    destruct o1, o3, H0 as [Ha Hb]; simpl in *.
    destruct agree_car, agree_car0; try discriminate; simpl in *.
    destruct (Ha o0) as [x []]; first constructor.
    lapply (Hvalid _ o1 H0); last constructor.
    rewrite <- discrete_iff, leibniz_equiv_iff in H2 |- * by auto; intro; subst; tauto.
  - intros [[[]|] [|]]; rewrite /= ?snap_valid_eq ?snap_validN_eq /=; split; intros;
      rewrite ?cmra_included_includedN ?cmra_valid_validN; naive_solver eauto using O.
  - intros n [[[]|] [|]]; rewrite !snap_validN_eq /=;
      naive_solver eauto using cmra_includedN_S, cmra_validN_S.
  - split; simpl; by rewrite assoc.
  - by split; simpl; rewrite comm.
  - destruct x; unfold core, op; simpl.
    unfold snap_op; simpl.
    rewrite left_id lift_join_idemp; auto.
  - by split; simpl; rewrite ?cmra_core_idemp.
  - intros [] [] [[] []]; simpl in *.
    unfold core; simpl.
    exists (Snap None snapshot2); unfold op; simpl; unfold snap_op; simpl.
    rewrite left_id H0; auto.
  - intros n [[[]|] b1] [[[]|] b2]; rewrite snap_validN_eq; simpl;
      destruct (lift_join b1 b2) eqn: Hj; try contradiction; destruct b1, b2; auto; try done.
    + assert (hd ε (agree_car a ++ agree_car a0) = hd ε (agree_car a)) as Heq.
      { destruct a; simpl in *.
        destruct agree_car; first discriminate; auto. }
      intros (? & ? & ?); repeat split; rewrite <- ?Heq; try naive_solver eauto
        using cmra_validN_op_l, cmra_validN_includedN.
      etransitivity; eauto.
      eapply join_le; eauto.
    + intros (? & ? & ?); repeat split; auto.
      etransitivity; eauto.
      eapply join_le; eauto.
  - intros n x y1 y2 ? [??]; simpl in *.
    destruct (cmra_extend n (master x) (master y1)
      (master y2)) as (ea1&ea2&?&?&?); auto using master_validN.
    rewrite <- discrete_iff, leibniz_equiv_iff in H1; last apply _.
    exists (Snap ea1 (snapshot y1)), (Snap ea2 (snapshot y2)); repeat split; try done.
    rewrite H1; auto.
Qed.
Canonical Structure snapR := CmraT (snap A) snap_cmra_mixin.

Instance snap_empty : Unit (snap A) := Snapshot ε.
Lemma snap_ucmra_mixin : UcmraMixin (snap A).
Proof.
  split; simpl.
  - by rewrite snap_valid_eq /=.
  - intros x; constructor; rewrite /= ?left_id; try done.
    destruct (snapshot x); auto.
    rewrite join_unit; auto.
  - auto.
Qed.
Canonical Structure snapUR := UcmraT (snap A) snap_ucmra_mixin.

Global Instance snap_frag_persistent a : CoreId (Snapshot a).
Proof. do 2 constructor; simpl; auto. Qed.

(** Internalized properties *)
Lemma snap_equivI {M} (x y : snap A) :
  x ≡ y ⊣⊢ (master x ≡ master y ∧ snapshot x ≡ snapshot y : uPred M).
Proof. by uPred.unseal. Qed.
Lemma snap_validI {M} (x : snap A) :
  ✓ x ⊣⊢ (match snapshot x with None => False | Some b => match master x with
          | Some (k, a) => ✓ k ∧ bi_pure (R b (hd ε (agree_car a))) ∧ ✓ a
          | None => True
          end end : uPred M).
Proof. uPred.unseal. by destruct x as [[[]|] [|]]. Qed.

Lemma snap_frag_op a b c : join a b = Some c -> Snapshot c = Snapshot a ⋅ Snapshot b.
Proof. intro; unfold op; simpl; unfold snap_op; simpl.
  rewrite H; auto.
Qed.
Lemma snap_frag_mono a b : R a b → Snapshot a ≼ Snapshot b.
Proof. intro. exists (Snapshot b); split; auto.
  unfold op; simpl.
  rewrite le_join; auto.
Qed.

(*Global Instance snap_frag_sep_homomorphism :
  MonoidHomomorphism op op (≡) (Snap None).
Proof. by split; [split; try apply _|]. Qed.*)

Lemma master_op : forall p q a, Master p a ⋅ Master q a ≡ Master (p ⋅ q) a.
Proof.
  intros; split; simpl.
  - unfold op, cmra_op; simpl.
    unfold op, cmra_op; simpl.
    unfold prod_op; simpl.
    rewrite agree_idemp; auto.
  - by rewrite join_unit.
Qed.

Lemma master_op_inv p a q b : ✓ (Master p a ⋅ Master q b) → a ≡ b.
Proof.
  unfold op; simpl; unfold snap_op; simpl.
  rewrite join_unit; intros (? & ? & ?).
  apply agree_op_inv'; auto.
Qed.
Lemma master_op_invL `{!LeibnizEquiv A} q a p b : ✓ (Master p a ⋅ Master q b) → a = b.
Proof. by intros ?%master_op_inv%leibniz_equiv. Qed.
Lemma master_op_invN n p a q b : ✓{n} (Master p a ⋅ Master q b) → a ≡{n}≡ b.
Proof.
  unfold op; simpl; unfold snap_op; simpl.
  rewrite join_unit; intros (? & ? & ?).
  apply agree_op_invN, to_agree_injN in H1; auto.
Qed.

Lemma snap_both_op k a b : Snap (Some (k, to_agree a)) (Some b) ≡ Master k a ⋅ Snapshot b.
Proof. by rewrite /op /snap_op /= join_unit. Qed.
Lemma snap_snap_valid k a : ✓ k → ✓ (Master k a).
Proof using L J D. intros; repeat split; simpl; eauto using ucmra_unit_leastN. apply unit_min. Qed.

Lemma make_snap q a : Master q a ~~> Master q a ⋅ Snapshot a.
Proof.
  intros; apply cmra_total_update; intros.
  destruct z as [[[]|] bf1]; hnf; simpl in *; rewrite join_unit; destruct bf1; try contradiction;
    hnf in H; simpl in *.
  - rewrite join_unit in H.
    rewrite join_comm le_join; repeat split; try tauto; done.
  - rewrite join_unit in H.
    rewrite join_comm le_join; repeat split; try tauto; done.
Qed.

Lemma snap_update a b a' :
  R a a' → Master 1 a ⋅ Snapshot b ~~> Master 1 a' ⋅ Snapshot a'.
Proof.
  intros Ha' Hlt; apply cmra_total_update; intros.
  destruct z as [[[]|] bf1]; hnf; hnf in H; simpl in *; rewrite !join_unit in H |- *;
    destruct bf1; try contradiction; simpl in *; destruct (join b o) eqn: Hj; try contradiction;
    destruct H as (? & ? & ?).
  { apply Qp_not_plus_q_ge_1 in H; contradiction. }
  rewrite join_comm le_join; first done.
  etransitivity; eauto.
  etransitivity; eauto.
  eapply join_le; rewrite join_comm; eauto.
Qed.

Lemma master_update a b a' :
  R a a' → Master 1 a ~~> Master 1 a'.
Proof.
  intros Ha' Hlt; apply cmra_total_update; intros.
  destruct z as [[[]|] bf1]; hnf; hnf in H; simpl in *;
    destruct bf1; try contradiction; simpl in *; rewrite !join_unit in H |- *; destruct H as (? & ? & ?).
  { apply Qp_not_plus_q_ge_1 in H; contradiction. }
  split; last split; auto.
  etransitivity; eauto.
Qed.

(*Lemma snap_update_dealloc a b a' : (a,b) ~l~> (a',ε) → ● a ⋅ ◯ b ~~> ● a'.
Proof. intros. rewrite -(right_id _ _ (● a')). by apply snap_update. Qed.*)

(*Lemma snap_local_update (a b0 b1 a' b0' b1': A) :
  (b0, b1) ~l~> (b0', b1') → b0' ≼ a' → ✓ a' →
  (● a ⋅ ◯ b0, ● a ⋅ ◯ b1) ~l~> (● a' ⋅ ◯ b0', ● a' ⋅ ◯ b1').
Proof.
  rewrite !local_update_unital=> Hup ? ? n /=.
  move=> [[[ac|]|] bc] /snap_validN_2 [Le Val] [] /=;
    inversion_clear 1 as [?? Ha|]; inversion_clear Ha. (* need setoid_discriminate! *)
  rewrite !left_id=> ?.
  destruct (Hup n bc) as [Hval' Heq]; eauto using cmra_validN_includedN.
  rewrite -!snap_both_op snap_validN_eq /=.
  split_and!; [by apply cmra_included_includedN|by apply cmra_valid_validN|done].
Qed.*)

End cmra.

Arguments snapR {_} {_} {_} {_} {_} J.
Arguments snapUR {_} {_} {_} {_} {_} J.

(* Proof mode class instances *)
(*Instance is_op_snap_frag {A : ofeT} (a b1 b2 : A) :
  IsOp a b1 b2 → IsOp' (Snapshot a) (Snapshot b1) (Snapshot b2).
Proof. done. Qed.*)

(* Functor *)
Definition snap_map {A B} (f : A → B) (x : snap A) : snap B :=
  Snap (prod_map id (agree_map f) <$> master x) (f <$> (snapshot x)).
Lemma snap_map_id {A} (x : snap A) : snap_map id x = x.
Proof. destruct x as [[[]|] [|]]; unfold snap_map; simpl; try done; rewrite agree_map_id; auto.
Qed.
Lemma snap_map_compose {A B C} (f : A → B) (g : B → C) (x : snap A) :
  snap_map (g ∘ f) x = snap_map g (snap_map f x).
Proof. destruct x as [[[]|] [|]]; unfold snap_map; simpl; try done;
  rewrite agree_map_compose; auto. Qed.
Lemma snap_map_ext {A B : ofeT} (f g : A → B) {_ : NonExpansive f} x :
  (∀ x, f x ≡ g x) → snap_map f x ≡ snap_map g x.
Proof.
  constructor; simpl; auto.
  apply option_fmap_equiv_ext=> a.
  unfold prod_map; simpl.
  by rewrite agree_map_ext.
  apply option_fmap_equiv_ext; auto.
Qed.
Instance snap_map_ne {A B : ofeT} (f : A -> B) {NE : NonExpansive f} : NonExpansive (snap_map f).
Proof.
  intros n x y [H H']. split; simpl.
  - apply option_fmap_ne; [|done].
    apply prod_map_ne; [done|].
    apply agree_map_ne; done.
  - rewrite H'; done.
Qed.
Definition snapC_map {A B} (f : A -n> B) : snapC A -n> snapC B :=
  CofeMor (snap_map f).
Lemma snapC_map_ne A B : NonExpansive (@snapC_map A B).
Proof. intros n f f' Hf [[[]|] [|]]; do 2 constructor; try naive_solver; simpl.
  constructor; auto. apply agreeC_map_ne; auto.
  constructor; auto. apply agreeC_map_ne; auto.
Qed.

From iris.proofmode Require Import tactics.
From iris.base_logic Require Import own.
From igps Require Import weakestpre.

Lemma own_snap_ord : forall `{J : Join} Σ {sG : inG Σ (snapR J)} g q a b,
  ⎡own g (Master q a)⎤ -∗ ⎡own g (Snapshot b)⎤ -∗ ⌜R b a⌝ : vProp Σ.
Proof.
  intros.
  iIntros "H1 H2".
  iPoseProof (own_valid_2 with "H1 H2") as "Hvalid".
  rewrite snap_validI; simpl.
  rewrite join_unit.
  by iDestruct "Hvalid" as "(? & % & ?)".
Qed.

Lemma own_master_eq : forall `{J : Join} Σ {sG : inG Σ (snapR J)} g q r a b,
  ⎡own g (Master q a)⎤ -∗ ⎡own g (Master r b)⎤ -∗ ⌜a = b⌝ : vProp Σ.
Proof.
  intros.
  iIntros "H1 H2".
  iPoseProof (own_valid_2 with "H1 H2") as "Hvalid".
  rewrite snap_validI; simpl.
  rewrite join_unit.
  iDestruct "Hvalid" as "(? & % & Hagree)".
  rewrite agree_validI agree_equivI.
  iDestruct "Hagree" as %?; auto.
Qed.

Lemma half_join : (1/2 ⋅ 1/2 = 1)%Qp.
Proof.
  unfold op, cmra_op; simpl.
  unfold frac_op.
  apply Qp_div_2.
Qed.