From iris.program_logic Require Export weakestpre.
From iris.proofmode Require Import tactics.

From igps Require Import notation malloc repeat protocols.
From igps.gps Require Import shared plain.

Import uPred.

(** Basic Treiber Stack in Release-Acquire **)

Section TreiberStack.

  Local Notation next := 0%Z.
  Local Notation data := 1%Z.

  Definition newStack : base.val :=
    λ: <>,
      let: "s" := Alloc in
      ["s"]_na <- #0  ;; "s".

  Definition tryPush : base.val :=
    λ: "s" "x",
      let: "n" := malloc (#2) in
      ["n" + #data]_na <- "x" ;;
      let: "h" := ["s"]_at in
      ["n" + #next]_na <- "h";;
      CAS "s" "h" ((int)"n").

  Definition tryPop : base.val :=
    λ: "s",
      let: "h" := ["s"]_at in
      if: "h" = #0
      then #0
      else
        let: "n" := [((loc)"h") + #next]_na in
        if: (CAS "s" "h" "n")
        then [((loc)"h") + #data]_na
        else #0.

  Section Interpretation.
    Context `{fG : foundationG Σ}
            `{gpG: !gpsG Σ unitProtocol _}
            `{persG : persistorG Σ}
            `{cG : cancelable_invariants.cinvG Σ}.
    Set Default Proof Using "Type".

    Definition Reachable' P: (Z -c> list Z -c> vProp Σ)
                             -c> Z -c> list Z -c> vProp Σ
    := (fun F n A =>
        match A with
        | nil => ⌜n = 0⌝%Z
        | v::A' => ⌜0 < n⌝%Z ∗ Z.to_pos (n + data) ↦ v ∗ P v ∗
                    ∃ q n', Z.to_pos (n + next) ↦{ q } n' ∗ ▷ F n' A'
        end)%I.

    Instance Reachable'_inhabited: Inhabited (Z -c> list Z -c> vProp Σ)
      := populate (λ _ _, True)%I.

    Instance Reachable'_contractive P:
        Contractive (Reachable' P).
    Proof.
      intros ? ? ? H ? A.
      unfold Reachable'.
      destruct A as [|v A]; [done|].
      repeat (apply bi.sep_ne; [done|]).
      repeat (apply bi.exist_ne => ?).
      apply bi.sep_ne; [done|].
      apply later_contractive. destruct n => //. by apply (H).
    Qed.

    Definition Reachable P := fixpoint (Reachable' P).

    Definition ReachableD': (Z -c> list Z -c> vProp Σ)
                             -c> Z -c> list Z -c> vProp Σ
    := (fun F n A =>
        match A with
        | nil => ⌜n = 0⌝%Z
        | n'::A' => ⌜0 < n⌝%Z ∗ ∃ q, Z.to_pos (n + next) ↦{ q } n' ∗ ▷ F n' A'
        end)%I.

    Instance ReachableD'_contractive:
        Contractive (ReachableD').
    Proof.
      intros ? ? ? H ? A.
      destruct A as [|v A]; [done|].
      repeat (apply bi.sep_ne; [done|]).
      apply bi.exist_ne => ?.
      apply bi.sep_ne; [done|].
      apply later_contractive. destruct n => //. by apply (H).
    Qed.

    Definition ReachableD := fixpoint (ReachableD').

    Definition Head P : interpC Σ unitProtocol
      := (λ b _ _ n, if b then ∃ A, Reachable P n A else ∃ A, ReachableD n A)%I.

    Lemma Head_dup P l s n:
      Head P false l s n -∗ Head P false l s n ∗ Head P false l s n.
    Proof.
      iIntros "Head".
      iDestruct "Head" as (A) "Head".
      rewrite /ReachableD (fixpoint_unfold (ReachableD') _ _).
      iInduction A as [|n' A'] "IH" forall (n).
      - iDestruct "Head" as % ->.
        iSplitL; iExists []; by rewrite /ReachableD (fixpoint_unfold (ReachableD') _ _).
      - iDestruct "Head" as "(% & Head)".
        iDestruct "Head" as (q) "(next & Head)".
        iAssert (▷ ((Head P false l s n') ∗ (Head P false l s n')))%I
          with "[Head]" as "(Hn1 & Hn2)".
        { iNext. rewrite (fixpoint_unfold (ReachableD') _ _).
          iApply ("IH" with "Head"). }
        iDestruct (na_frac_split_2 with "next") as "[NA1 NA2]".
        iSplitL "Hn1 NA1";
          [iDestruct "Hn1" as (A1) "Hn1"; iExists (n' :: A1)
           | iDestruct "Hn2" as (A2) "Hn2"; iExists (n' :: A2)];
           rewrite /ReachableD !(fixpoint_unfold (ReachableD') _ _);
           (iSplitL ""; first done); iExists (q/2)%Qp;
           [iFrame "NA1"| iFrame "NA2"];
           iNext; by rewrite (fixpoint_unfold (ReachableD') _ _).
    Qed.

    Lemma Head_dup_main P l s n:
      Head P true l s n -∗ Head P true l s n ∗ Head P false l s n.
    Proof.
      iIntros "Head".
      iDestruct "Head" as (A) "Head".
      rewrite /Reachable (fixpoint_unfold (Reachable' P) _ _).
      iInduction A as [|n' A'] "IH" forall (n).
      - iDestruct "Head" as "#Head".
        iSplitL; iExists [].
        + by rewrite /Reachable' /Reachable (fixpoint_unfold (Reachable' P) _ _).
        + by rewrite /ReachableD (fixpoint_unfold (ReachableD') _ _).
      - iDestruct "Head" as "(% & od & P & Head)".
        iDestruct "Head" as (q h') "(next & Head)".
        iAssert (▷ ((Head P true l s h') ∗ (Head P false l s h')))%I
          with "[Head]" as "(Hn1 & Hn2)".
        { iNext. rewrite (fixpoint_unfold (Reachable' P) _ _).
          iApply ("IH" with "Head"). }
        iDestruct (na_frac_split_2 with "next") as "[NA1 NA2]".
        iSplitL "Hn1 NA1 od P".
        + iDestruct "Hn1" as (A1) "Hn1"; iExists (n' :: A1).
          rewrite /Reachable !(fixpoint_unfold (Reachable' P) _ _).
          iSplitL ""; first done. iFrame "od P".
          iExists _, _. iFrame "NA1".
          iNext. by rewrite (fixpoint_unfold (Reachable' P) _ _).
        + iDestruct "Hn2" as (A2) "Hn2"; iExists (h' :: A2).
          rewrite /ReachableD !(fixpoint_unfold (ReachableD') _ _).
          iSplitL ""; first done.
          iExists _. iFrame "NA2".
          iNext; by rewrite (fixpoint_unfold (ReachableD') _ _).
    Qed.
  End Interpretation.

  Section proof.
    Context `{fG : foundationG Σ}
            `{gpG: !gpsG Σ unitProtocol _}
            `{persG : persistorG Σ}
            `{agreeG : !gps_agreeG Σ unitProtocol}
            `{cG : cancelable_invariants.cinvG Σ}.
    Set Default Proof Using "Type".

    Context (P : Z → vProp Σ).

    Definition Stack s := [PP s in () | Head P]%I.

    Lemma newStack_spec:
      {{{ ⎡PSCtx⎤ ∗ ⎡PersistorCtx⎤ }}}
          (newStack #())
      {{{ (s: loc), RET #s; Stack s }}}.
    Proof.
      intros. iIntros "[#kI #kP] Post".
      wp_lam. wp_bind Alloc.
      iApply (alloc with "kI"); [done|].
      iNext. iIntros (s) "os".
      wp_seq. wp_bind ([_]_na <- _)%E.
      iApply (GPS_PP_Init_default (Head P) () with "[$kI $kP] [$os]");
        [done|done|..].
      { iSplitL; iExists [].
        - by rewrite /Reachable (fixpoint_unfold (Reachable' P) _ _).
        - by rewrite /ReachableD (fixpoint_unfold (ReachableD') _ _). }
      iNext. iIntros "Head".
      wp_seq.
      by iApply ("Post" with "Head").
    Qed.

    Lemma tryPush_spec s x:
      {{{ ⎡PSCtx⎤ ∗ Stack s ∗ P x}}}
        tryPush #s #x
      {{{ (z: Z), RET #z; ⌜z ≠ 0⌝%Z ∨ P x }}}.
    Proof.
      intros. iIntros "[#kI [Stack P]] Post".
      wp_lam. wp_let.
      wp_bind (malloc _).

      iApply (wp_mask_mono); first auto.
      iApply (malloc_spec); [omega|done|done|].
      iNext. iIntros (n) "oLs".
      rewrite big_sepL_cons big_sepL_singleton.
      iDestruct "oLs" as "(ol & od)".
      wp_let. wp_bind ([_]_na <- _)%E. wp_op.
      iApply (na_write with "[$kI $od]"); [done|].
      iNext. iIntros "od".
      wp_seq. wp_bind ([_]_at)%E.
      iApply (GPS_PP_Read (p := ()) (Head P) True%I (λ _ _, True)%I
          with "kI [$Stack]");
        [done|done|..].
      { iSplitL.
        - iIntros (?) "_". iLeft. by iIntros "!>" (?) "_".
        - rewrite monPred_objectively_unfold. setoid_rewrite <-Head_dup.
          by iIntros "!>" (?????) "[_ $]". }

      iNext. iIntros ([] h) "(_ & Head & _)".
      wp_seq. wp_bind ([_]_na <- _)%E. wp_op.
      iApply (na_write_frac_1 with "[$kI $ol]"); [done|].
      iNext. iIntros "ol".
      wp_seq. wp_op.

      set P' := (P x ∗ ZplusPos 1 n ↦ x ∗ ZplusPos 0 n ↦{1} h)%I.
      iApply (GPS_PP_CAS (p := ()) (Head P) P' (λ _, True)%I (λ _ _, P')
                with "kI [P od ol $Head]"); [done|done|..].
      { iSplitL ""; last iSplitL ""; last iSplitL "".
        - iNext. iIntros ([]) "(_ & Head & P & od & ol)".
          iExists (). do 2 (iSplitL ""; first done).
          iModIntro. rewrite -bi.later_sep. iApply Head_dup_main.
          iDestruct "Head" as (A) "Head".
          iExists (x :: A).
          rewrite /Reachable !(fixpoint_unfold (Reachable' P) _ _).
          iSplitL ""; first (iPureIntro; lia). iFrame "P".
          iSplitL "od".
          + rewrite (_: Z.to_pos (Z.pos n + 1) = ZplusPos 1 n); first done.
            rewrite /ZplusPos. f_equal. omega.
          + iExists _, _. iFrame "ol". iNext.
            by rewrite /Reachable' (fixpoint_unfold (Reachable' P) _ _).
        - iNext. by iIntros (??) "(_ & _ & _ & ?)".
        - rewrite monPred_objectively_unfold. setoid_rewrite <-Head_dup.
          iIntros (?????) "!> [_ $] //".
        - by iFrame. }

      iNext. iIntros ([] [] _) "(_ & Head & IF)".
      - iApply ("Post" with "[]"); by iLeft.
      - iDestruct "IF" as "[P ?]".
        iApply ("Post" with "[P]"); by iRight.
    Qed.

    Lemma tryPop_spec s:
      {{{ ⎡PSCtx⎤ ∗ Stack s }}}
        tryPop #s
      {{{ (x: Z), RET #x; ⌜x = 0⌝%Z ∨ P x }}}.
    Proof.
      intros. iIntros "[#kI Stack] Post".
      wp_lam. wp_bind ([_]_at)%E.

      set Q : pr_state unitProtocol → Z → vProp Σ :=
        λ _ h, (⌜h ≠ 0⌝%Z →
                ⌜0 < h⌝%Z ∗ ∃ q n, (Z.to_pos (h + next)) ↦{q} n)%I.
      iApply (GPS_PP_Read (p := ()) (Head P) True%I Q with "kI [$Stack]");
        [done|done|..].
      { iSplitL.
        - iIntros (?) "_". iLeft. iIntros "!>" (h) "[Head _]".
          iDestruct "Head" as (A) "Head".
          iModIntro. iIntros "%".
          rewrite /ReachableD (fixpoint_unfold (ReachableD') _ _).
          destruct A as [|h' A'].
          + by iDestruct "Head" as "%".
          + iDestruct "Head" as "[$ Head]".
            iDestruct "Head" as (q) "[Head _]". by iExists _, _.
        - rewrite monPred_objectively_unfold. setoid_rewrite <- Head_dup.
          iIntros "!>" (?????) "[_ $] //". }

      iNext. iIntros ([] h) "(_ & Head & Q)".
      wp_seq. wp_op.
      case_bool_decide.
      - wp_if_true.
        iApply ("Post"); by iLeft.
      - wp_if_false. wp_bind ([_]_na)%E.
        wp_op. wp_op.
        iDestruct ("Q" with "[%]") as "[% oh]"; [done|].
        iDestruct "oh" as (q n) "oh".
        iApply (na_read_frac with "[$kI oh]"); [done|done|..].
        { rewrite (_: ZplusPos 0 (Z.to_pos h) = Z.to_pos (h + 0)); first done.
          rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|done]. }
        iNext. iIntros (z) "(% & oh)". subst z.

        wp_let.  wp_bind (CAS _ _ _).

        set P' := (ZplusPos 0 (Z.to_pos h) ↦{q} n)%I.
        set Q': pr_state unitProtocol → _ :=
          λ _, (∃ v, (ZplusPos data (Z.to_pos h)) ↦ v ∗ P v)%I.
        iApply (GPS_PP_CAS (p := ()) (Head P) P' Q' (λ _ _, True)%I
                with "kI [$oh $Head]"); [done|done|..].
        { iSplitL; last iSplitL.
          - iIntros "!>" ([]) "(_ & Head & oh)".
            iExists (). iSplitL ""; first done.
            iDestruct "Head" as (A) "Head".
            rewrite /Reachable (fixpoint_unfold (Reachable' P) _ _).
            destruct A as [|v A'].
            + by iDestruct "Head" as "%".
            + iDestruct "Head" as "(_ & od & P & Head)".
              iDestruct "Head" as (q' n') "[oh' Head]".
              rewrite /P'.
              iDestruct (na_frac_agree with "[$oh' oh]") as "%".
              { rewrite (_: ZplusPos 0 (Z.to_pos h) = Z.to_pos (h + 0));
                  first done.
                rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|done]. }
              subst n'. iModIntro.
              iSplitL "od P".
              * iExists _. iFrame "P".
                rewrite (_: ZplusPos 1 (Z.to_pos h) = Z.to_pos (h + 1));
                  first done.
                rewrite /ZplusPos. f_equal. rewrite Z2Pos.id; [omega|done].
              * rewrite -bi.later_sep. iNext. iApply Head_dup_main.
                iExists A'.
                by rewrite /Reachable (fixpoint_unfold (Reachable' P) _ _).
          - iNext. by iIntros.
          - rewrite monPred_objectively_unfold. setoid_rewrite <- Head_dup.
            iIntros "!>" (?????) "[_ $] //". }

        iNext. iIntros ([] [] _) "(_ & Head & IF)".
        + wp_if_true.
          wp_op. wp_op.
          iDestruct "IF" as (v) "[od P]".
          iApply (na_read with "[$kI $od]"); [done|].
          iNext. iIntros (z) "(% & od)". subst z.
          iApply ("Post" with "[P]"); by iRight.
        + wp_if_false.
          iApply ("Post"); by iLeft.
    Qed.

  End proof.
End TreiberStack.
