From iris.program_logic Require Import ectx_language ectxi_language language.
From stdpp Require Export strings gmap.
From iris.algebra Require Import ofe.

From igps Require Export machine.

Open Scope Z_scope.

(** Expressions and vals. *)

Inductive base_lit : Set :=
  | LitInt (n : Z) | LitUnit | LitLoc (l : loc).
Inductive un_op : Set :=
  | NegOp | MinusUnOp | CastIntOp | CastLocOp.
Inductive bin_op : Set :=
  | PlusOp | MinusOp | TimesOp | ModOp | LeOp | LtOp | EqOp.

Inductive binder := BAnon | BNamed : string → binder.
Delimit Scope binder_scope with bind.
Bind Scope binder_scope with binder.
Definition cons_binder (mx : binder) (X : list string) : list string :=
  match mx with BAnon => X | BNamed x => x :: X end.
Infix ":b:" := cons_binder (at level 60, right associativity).
Instance binder_dec_eq (x1 x2 : binder) : Decision (x1 = x2).
Proof.
  (* infinite loop yeah yeah yeah *)
  (* solve_decision. *)
  destruct x1,x2; try (by left); try by right.
  case: (decide (s = s0)) => [->|]; [by left|by right => [[]]].
Defined.

Instance set_unfold_cons_binder x mx X P :
  SetUnfold (x ∈ X) P → SetUnfold (x ∈ mx :b: X) (BNamed x = mx ∨ P).
Proof.
  constructor. rewrite -(set_unfold (x ∈ X) P).
  destruct mx; rewrite /= ?elem_of_cons; abstract naive_solver.
Qed.

Inductive expr :=
  (* Base lambda calculus *)
  | Var (x : string)
  | Rec (f x : binder) (e : expr)
  | App (e1 e2 : expr)
  (* Base types and their operations *)
  | Lit (l : base_lit)
  | UnOp (op : un_op) (e : expr)
  | BinOp (op : bin_op) (e1 e2 : expr)
  | If (e0 e1 e2 : expr)
  (* Concurrency *)
  | Fork (e : expr)
  (* Memory *)
  | Alloc
  | Dealloc (e : expr)
  | Load (acc : access) (e : expr)
  | Store (acc : access) (e1 : expr) (e2 : expr)
  | CAS (e0 : expr) (e1 : expr) (e2 : expr)
  | FAI (C: positive) (e0 : expr).


Fixpoint is_closed (X : list string) (e : expr) : bool :=
  match e with
  | Var x => bool_decide (x ∈ X)
  | Rec f x e => is_closed (f :b: x :b: X) e
  | Lit _ | Alloc => true
  | UnOp _ e | Fork e | Load _ e | FAI _ e | Dealloc e =>
      is_closed X e
  | App e1 e2 | BinOp _ e1 e2 | Store _ e1 e2 =>
      is_closed X e1 && is_closed X e2
  | If e0 e1 e2 | CAS e0 e1 e2 =>
      is_closed X e0 && is_closed X e1 && is_closed X e2
  end.

Class Closed (X : list string) (e : expr) := closed : is_closed X e.
Instance closed_proof_irrel env e : ProofIrrel (Closed env e).
Proof. rewrite /Closed. apply _. Qed.
Instance closed_decision env e : Decision (Closed env e).
Proof. rewrite /Closed. apply _. Qed.

Inductive val :=
  | RecV (f x : binder) (e : expr) `{!Closed (f :b: x :b: []) e}
  | LitV (l : base_lit).

Bind Scope val_scope with val.
Delimit Scope val_scope with V.

Fixpoint of_val (v : val) : expr :=
  match v with
  | RecV f x e _ => Rec f x e
  | LitV l => Lit l
  end.

Fixpoint to_val (e : expr) : option val :=
  match e with
  | Rec f x e =>
      if decide (Closed (f :b: x :b: []) e) then Some (RecV f x e) else None
  | Lit l => Some (LitV l)
  | _ => None
  end.

(** Evaluation contexts *)
Inductive ectx_item :=
  | AppLCtx (e2 : expr)
  | AppRCtx (v1 : val)
  | UnOpCtx (op : un_op)
  | BinOpLCtx (op : bin_op) (e2 : expr)
  | BinOpRCtx (op : bin_op) (v1 : val)
  | IfCtx (e1 e2 : expr)
  | DeallocCtx
  | LoadCtx (acc : access)
  | StoreLCtx (acc : access) (e2 : expr)
  | StoreRCtx (acc : access) (v1 : val)
  | CasLCtx (e1 : expr)  (e2 : expr)
  | CasMCtx (v0 : val) (e2 : expr)
  | CasRCtx (v0 : val) (v1 : val)
  | FaiCtx (C: positive).

Definition fill_item (Ki : ectx_item) (e : expr) : expr :=
  match Ki with
  | AppLCtx e2 => App e e2
  | AppRCtx v1 => App (of_val v1) e
  | UnOpCtx op => UnOp op e
  | BinOpLCtx op e2 => BinOp op e e2
  | BinOpRCtx op v1 => BinOp op (of_val v1) e
  | IfCtx e1 e2 => If e e1 e2
  | DeallocCtx => Dealloc e
  | LoadCtx acc => Load acc e
  | StoreLCtx acc e2 => Store acc e e2
  | StoreRCtx acc v1 => Store acc (of_val v1) e
  | CasLCtx e1 e2 => CAS e e1 e2
  | CasMCtx v0 e2 => CAS (of_val v0) e e2
  | CasRCtx v0 v1 => CAS (of_val v0) (of_val v1) e
  | FaiCtx C => FAI C e
  end.

Definition fill (K : list ectx_item) (e : expr) : expr := foldl (flip fill_item) e K.

(** Substitution *)
Fixpoint subst (x : string) (es : expr) (e : expr)  : expr :=
  match e with
  | Var y => if decide (x = y) then es else Var y
  | Rec f y e =>
      Rec f y $ if decide (BNamed x ≠ f ∧ BNamed x ≠ y) then subst x es e else e
  | App e1 e2 => App (subst x es e1) (subst x es e2)
  | Lit l => Lit l
  | UnOp op e => UnOp op (subst x es e)
  | BinOp op e1 e2 => BinOp op (subst x es e1) (subst x es e2)
  | If e0 e1 e2 => If (subst x es e0) (subst x es e1) (subst x es e2)
  | Fork e => Fork (subst x es e)
  | Alloc => Alloc
  | Dealloc e => Dealloc (subst x es e)
  | Load acc e => Load acc (subst x es e)
  | Store acc e1 e2 => Store acc (subst x es e1) (subst x es e2)
  | CAS e0 e1 e2 => CAS (subst x es e0) (subst x es e1) (subst x es e2)
  | FAI C e => FAI C (subst x es e)
  end.

Definition subst' (mx : binder) (es : expr) : expr → expr :=
  match mx with BNamed x => subst x es | BAnon => id end.

(** The stepping relation *)
Definition un_op_eval (op : un_op) (l : val) : option val :=
  match op, l with
  | NegOp, LitV (LitInt n) => Some $ LitV $ LitInt (if decide (n = 0) then 1 else 0)
  | MinusUnOp, LitV (LitInt n) => Some $ LitV $ LitInt (- n)
  | CastIntOp, LitV (LitLoc n) => Some $ LitV $ LitInt (Z.pos n)
  | CastLocOp, LitV (LitInt n) => Some $ LitV $ LitLoc $ Z.to_pos n
  | _, _ => None
  end.

Definition ZplusPos (n: Z) (p: positive) : positive :=
  Z.to_pos (n + Zpos p).

Definition Z_of_bool (b : bool) : Z :=
  if b then 1 else 0.

Definition lit_of_bool (b : bool) : base_lit :=
  LitInt $ Z_of_bool b.

Definition bin_op_eval (op : bin_op) (l1 l2 : val) : option val :=
  match op, l1, l2 with
  | PlusOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ LitInt (n1 + n2)
  | PlusOp, LitV (LitInt n), LitV (LitLoc l) => Some $ LitV $ LitLoc $ ZplusPos n l
  | PlusOp, LitV (LitLoc l), LitV (LitInt n) => Some $ LitV $ LitLoc $ ZplusPos n l
  | MinusOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ LitInt (n1 - n2)
  | TimesOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ LitInt (n1 * n2)
  | ModOp,  LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ LitInt (n1 `mod` n2)
  | LeOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ lit_of_bool $ bool_decide (n1 ≤ n2)
  | LtOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ lit_of_bool $ bool_decide (n1 < n2)
  | EqOp, LitV (LitInt n1), LitV (LitInt n2) => Some $ LitV $ lit_of_bool $ bool_decide (n1 = n2)
  | EqOp, LitV (LitLoc n1), LitV (LitLoc n2) => Some $ LitV $ lit_of_bool $ bool_decide (n1 = n2)
  | _, _, _ => None
  end.

Inductive head_step : expr → expr → Prop :=
  | BetaS f x e1 e2 v2 e' :
      to_val e2 = Some v2 →
      Closed (f :b: x :b: []) e1 →
      e' = subst' x (of_val v2) (subst' f (Rec f x e1) e1) →
      head_step (App (Rec f x e1) e2) e'
  | UnOpS op e v v' :
      to_val e = Some v →
      un_op_eval op v = Some v' →
      head_step (UnOp op e) (of_val v')
  | BinOpS op e1 e2 v1 v2 v':
      to_val e1 = Some v1 → to_val e2 = Some v2 →
      bin_op_eval op v1 v2 = Some v' →
      head_step (BinOp op e1 e2) (of_val v')
  | IfTrueS e1 e2 :
      head_step (If (Lit $ lit_of_bool true) e1 e2) e1
  | IfFalseS e1 e2 :
      head_step (If (Lit $ lit_of_bool false) e1 e2) e2.

Definition head_step' {state} : expr → state → expr → state → list expr → Prop := λ e1 _ e2 _ _, head_step e1 e2.

(** Basic properties about the language *)
Lemma to_of_val v : to_val (of_val v) = Some v.
Proof.
  by induction v; simplify_option_eq; repeat f_equal; try apply (proof_irrel _).
Qed.

Lemma of_to_val e v : to_val e = Some v → of_val v = e.
Proof.
  revert v; induction e; intros v ?; simplify_option_eq; auto with f_equal.
Qed.

Instance of_val_inj : Inj (=) (=) of_val.
Proof. by intros ?? Hv; apply (inj Some); rewrite -!to_of_val Hv. Qed.

Instance fill_item_inj Ki : Inj (=) (=) (fill_item Ki).
Proof. destruct Ki; intros ???; simplify_eq/=; auto with f_equal. Qed.

Lemma fill_item_val Ki e :
  is_Some (to_val (fill_item Ki e)) → is_Some (to_val e).
Proof. intros [v ?]. destruct Ki; simplify_option_eq; eauto. Qed.

Lemma val_stuck e1 e2 : head_step e1 e2 → to_val e1 = None.
Proof. destruct 1; abstract naive_solver. Qed.

Lemma head_ctx_step_val Ki e e2 :
  head_step (fill_item Ki e) e2 → is_Some (to_val e).
Proof. destruct Ki; inversion_clear 1; simplify_option_eq; by eauto. Qed.

Lemma fill_item_no_val_inj Ki1 Ki2 e1 e2 :
  to_val e1 = None → to_val e2 = None →
  fill_item Ki1 e1 = fill_item Ki2 e2 → Ki1 = Ki2.
Proof.
  destruct Ki1, Ki2; intros; try discriminate; simplify_eq/=;
    repeat match goal with
    | H : to_val (of_val _) = None |- _ => by rewrite to_of_val in H
    end; auto.
Qed.

(** Closed expressions *)
Lemma is_closed_weaken X Y e : is_closed X e → X ⊆ Y → is_closed Y e.
Proof. revert X Y; induction e; abstract naive_solver (eauto; set_solver). Qed.

Lemma is_closed_weaken_nil X e : is_closed [] e → is_closed X e.
Proof. intros. by apply is_closed_weaken with [], list_subseteq_nil. Qed.

Lemma is_closed_of_val X v : is_closed X (of_val v).
Proof. apply is_closed_weaken_nil. induction v; simpl; auto. Qed.

Lemma is_closed_to_val X e v : to_val e = Some v → is_closed X e.
Proof. intros Hev; rewrite -(of_to_val e v Hev); apply is_closed_of_val. Qed.

Lemma is_closed_subst X e x es : is_closed X e → x ∉ X → subst x es e = e.
Proof.
  revert X. induction e=> X /=; rewrite ?bool_decide_spec ?andb_True=> ??;
    repeat case_decide; simplify_eq/=; f_equal; intuition eauto with set_solver.
Qed.

Lemma is_closed_nil_subst e x es : is_closed [] e → subst x es e = e.
Proof. intros. apply is_closed_subst with []; set_solver. Qed.

(** Equality and other typeclass stuff *)
Instance base_lit0_dec_eq (l1 l2 : base_lit) : Decision (l1 = l2).
Proof. solve_decision. Defined.
Instance base_lit_dec_eq (l1 l2 : base_lit) : Decision (l1 = l2).
Proof.
  destruct l1, l2; try (by right); try (by left);
  eauto with typeclass_instances.
Defined.
Instance un_op_dec_eq (op1 op2 : un_op) : Decision (op1 = op2).
Proof. solve_decision. Defined.
Instance bin_op_dec_eq (op1 op2 : bin_op) : Decision (op1 = op2).
Proof. solve_decision. Defined.
Instance expr_dec_eq (e1 e2 : expr) : Decision (e1 = e2).
Proof. solve_decision. Defined.
Instance val_dec_eq (v1 v2 : val) : Decision (v1 = v2).
Proof.
  refine (cast_if (decide (of_val v1 = of_val v2))); abstract naive_solver.
Defined.

Instance expr_inhabited : Inhabited (expr) := populate (Lit LitUnit).
Instance val_inhabited : Inhabited val := populate (LitV LitUnit).

Canonical Structure stateC := leibnizC state.
Canonical Structure valC := leibnizC val.
Canonical Structure exprC := leibnizC expr.


Bind Scope expr_scope with expr.
Delimit Scope expr_scope with E.
Arguments Rec _ _ _%E.
Arguments App _%E _%E.
Arguments UnOp _ _%E.
Arguments BinOp _ _%E _%E.
Arguments If _%E _%E _%E.
Arguments Fork _%E.
Arguments Load _ _%E.
Arguments Store _ _%E _%E.
Arguments CAS _%E _%E _%E.
Arguments FAI _ _%E.
